# Store the version here so:
# 1) we don't load dependencies by storing it in __init__.py
# 2) we can import it in setup.py for the same reason
# 3) we can import it into your module module
# 0.7.4 - Correct class declaration to include object
# 0.7.5 - Added an alias for streamid to IEAN
# 0.7.6 - Added repr function to SimpleEthernet
# 0.8.0 - Lots of updates and documentation updates
# 0.8.1 - Added IENA-N and IENA-D packets
# 0.9.0 - Added NDP
# 0.9.2 - Added Ch10
# 0.9.4 - Previous chapter 10 was the UDP wrapper. Correct Chapter10 packet now added
# 0.9.5 - Added ARINC data packet
# 0.10.0 - Added iNET
# 0.10.1 - Added info to exception
# 0.10.2 - Fixed IP checksum bug
# 0.11.0 - Chapter10 endianness changes
# 0.11.1 - IENA-N and IENA-D throw exception if there are not integer number of D and N parameters in the payload
# 0.11.2 - ParserAligned throws exception if the number of quadbytes is illegal
# 0.12.0 - Added IENA-M and IENA-Q
# 0.12.1 - Changed pcap to an iterator model
# 0.12.2 - Fixed iNET byte order
# 0.12.3 - Fixed iNET byte order
# 0.12.4 - Changed INET to separate seconds and nanoseconds
# 0.13.0 - Added UART Payload for Chapter 10
# 0.13.1 - Bug fixes on UART for Chapter 10
# 0.13.2 - Added NPD Segments for RS232
# 0.13.3 - Parser aligned printout
# 0.14.0 - Made compatiable with python3
# 0.15.0 - Added ch7 and updated all unit test to pass in both py3 py2
# 0.15.1 - Added unpack to ParserAligned
# 0.15.2 - Fixd the ascii example
# 0.15.3 - Minor updates to Ch7 and unittest
# 0.15.4 - Pcap comment removal
# 0.15.5 - Handled padding for UART data words in chapter 10
# 0.15.6 - Optimisations on Golay encoding
# 0.15.7 - Minor updates
# 0.15.8 - Traffic generator update
# 0.15.10 - Traffic generator update
# 0.15.11 - Added FCS Support to ethernet packets
# 0.15.12 - Indentation bugfix
# 0.15.13 - Renamed PDFR to PTFR
# 0.15.14 - Massive perf improvements with Golay handling
# 0.15.15 - Added summary to the recorder
# 0.15.16 - Added summary to validation script
# 0.15.17 - Added IGMPv2 simplified packet generation
# 0.15.18 - Updated validate and pkt generation script
__version__ = '0.15.18'



#-------------------------------------------------------------------------------
# Name:        parser_aligned_pcap_example.py
# Purpose:
#
# Author:      Diarmuid
#
# Created:     19/12/2013
#
# Copyright 2014 Diarmuid Collins
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#-------------------------------------------------------------------------------

import socket,os,struct,sys
import argparse
import datetime, time

import AcraNetwork.iNetX as inetx
import AcraNetwork.Pcap as pcap
from AcraNetwork.SimpleEthernet import mactoreadable
import AcraNetwork.ParserAligned as ParserAligned

def main():

    #----------------------------------
    # Setup the command line parser
    #----------------------------------
    parser = argparse.ArgumentParser(description='Analyse a pcap file looking for iNetX parser aligned packets')
    parser.add_argument('--pcap',  required=True, action='append',  help='The dump pcap packet')
    args = parser.parse_args()

    #------------------------------------------------------------
    # Now read the input.
    #------------------------------------------------------------

    # The input will take multiple pcap files and loop through each

    for pcapfilename in args.pcap:
        try:
            pcapfile = pcap.Pcap(pcapfilename)
        except IOError:
            print("ERROR: File {} not found".format(pcapfilename))
            exit()


        packet_count = 1


        start_of_run = time.time() # benchmarking


        while True:
            try:

                # So we loop through the file one packet at a time. This will eventually return an
                # exception at the end of file so handle that when it occurs
                (eth_packet,ip_packet,udp_packet) = pcapfile._readNextUDPPacket()
                if udp_packet.isinetx: # This is a rough guess assuming the control word is 0x11000000
                    inetx_packet = inetx.iNetX()
                    inetx_packet.unpack(udp_packet.payload)
                    readablemac = mactoreadable(eth_packet.srcmac) # handy function to return the mac address in a readable format
                    output_format = "SRCMAC={:>20s} SRCIP={:>15s} DSTPORT={:5d} StreamID={:#5x} Sequence={:10d}"

                    # What string do we want outputted to the screen. The output format is defined in the coloredop class
                    outstring =output_format.format(readablemac ,ip_packet.srcip, udp_packet.dstport,inetx_packet.streamid,inetx_packet.sequence)
                    # Print out one line and the dropped packet info
                    print(outstring)

                    # We have a parser aligned block
                    if inetx_packet.streamid == 0x11121314: # This specific streamid is a parser aligned block
                        parser_aligned_packet = ParserAligned.ParserAlignedPacket()
                        # unpack the _payload as the parser data
                        parser_aligned_packet.unpack(inetx_packet.payload)
                        # Loop through all the blocks in the packet and spit them out
                        for pblock in parser_aligned_packet.parserblocks:
                            (payload_data,) =struct.unpack('>I',pblock.payload)
                            print("Quadb={:5} Msgcnt={:5} BusId={:4} Elapsed={:20}".format(pblock.quadbytes,pblock.messagecount,pblock.busid,pblock.elapsedtime,payload_data))


                packet_count += 1
            except NotImplementedError:
                # We received a packet that we don't care about. So skip silently
                packet_count += 1
                pass
            except IOError:
                # We are at the end of the file so lets jump to the next file
                print(( "End of file reached. Packets Per Second ={:5.1f}".format(packet_count/(time.time()-start_of_run))))
                break


if __name__ == '__main__':
    main()



"""
.. module:: NPD
    :platform: Unix, Windows
    :synopsis: Class to construct and de construct NPD Packets

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""
__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import struct
from socket import inet_aton,inet_ntoa


class NPDSegment(object):
    """
    NPD Payloads are split into segments. This class will pack and unpack segments
    
    :type timedelta: int
    :type segmentlen: int
    :type errorcode: int
    :type flags: int
    :type payload: str
    """

    NPD_SEGMENT_HDR_FORMAT = ">IHBB"
    NPD_SEGMENT_HDR_LEN = struct.calcsize(NPD_SEGMENT_HDR_FORMAT)

    def __init__(self):
        self.timedelta = None  #: The R-bit in the Flags field of the packet header dictates the format of this field.
        self.segmentlen = None #: The length of the segment header and data in bytes excluding padding.
        self.errorcode = None  #: This field has a zero value if there are no errors
        self.flags = None  #: [2:1] Fragmentation state flag
        self._payload = None  #: Payload of segment

    @property
    def payload(self):
        """
        Payload of segment
        :return: 
        """
        return self._payload

    @payload.setter
    def payload(self, buf):
        self._payload = buf
        self.segmentlen = len(self.payload) + NPDSegment.NPD_SEGMENT_HDR_LEN

    def unpack(self, buffer):
        """
        Unpack a string buffer into an NPD segment. Return the remaining buffer so that the next segment can iteratively
        be unpacked

        :param buffer: A string buffer representing an NPD segment
        :type buffer: str
        
        :rtype: str
        """
        (self.timedelta, self.segmentlen, self.errorcode, self.flags) = struct.unpack_from(NPDSegment.NPD_SEGMENT_HDR_FORMAT, buffer)
        self.payload = buffer[NPDSegment.NPD_SEGMENT_HDR_LEN:self.segmentlen]
        if self.segmentlen % 4  == 0:
            pad_len = 0
        else:
            pad_len = 4- (self.segmentlen % 4)

        return buffer[(self.segmentlen+pad_len):]

    def pack(self):
        """
        Pack the NPD object into a binary buffer

        :rtype: str 
        """
        if len(self.payload) % 4 == 0:
            pad = b""
        else:
            pad_len = 4 - len(self.payload) % 4
            pad = struct.pack(">B",0xFF) * pad_len

        hdr_pack = struct.pack(NPDSegment.NPD_SEGMENT_HDR_FORMAT, self.timedelta, self.segmentlen, self.errorcode, self.flags)

        return hdr_pack + self.payload + pad

    def __eq__(self, other):
        if not isinstance(other, NPDSegment):
            return False
        for attr in ["timedelta", "segmentlen", "errorcode", "flags", "payload"]:
            if getattr(other, attr) != getattr(self, attr):
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "NPD Segment. TimeDelta={} Segment Len={} ErrorCode={} Flags={:#0X}".format(
            self.timedelta, self.segmentlen, self.errorcode, self.flags)



class ACQSegment(NPDSegment):
    """
    PCM Segments
    """
    def __init__(self):
        NPDSegment.__init__(self)
        self.sfid = 0
        self.cal = 0
        self.words = []

    def unpack(self, buffer):
        remaining = NPDSegment.unpack(self, buffer)
        (self.sfid, _cal, reserved) = struct.unpack_from(">BBH", self.payload)
        self.cal = _cal >> 7
        len_words = (len(self.payload) - 4 ) / 2
        self.words = list(struct.unpack_from(">{}H".format(len_words), self.payload, 4))
        return remaining

    def __repr__(self):
        return "PCM NPD Segment. TimeDelta={} Segment Len={} ErrorCode={:#0X} Flags={:#0X} sfid={:#0X} " \
               "WordCnt={}" \
               "".format(self.timedelta, self.segmentlen, self.errorcode, self.flags, self.sfid, len(self.words))


class A429Segment(NPDSegment):
    pass


class RS232Segment(NPDSegment):

    BSL_CH0 = 0x0
    BSL_CH1 = 0x8000
    BSL_PAR_ERR = 0x4000
    BSL_TWO_STOP_BITS = 0x2000
    BSL_EVEN_PAR = 0x1000
    BSL_PARN_EN = 0x800
    BSL_8BIT = 0x0
    BSL_7BIT = 0x200
    BSL_6BIT = 0x400
    BSL_5BIT = 0x600
    BSL_PKT_SYNC_FIXED = 0x0
    BSL_PKT_SYNC_VAR = 0x80
    BSL_PKT_GAP = 0x100
    BSL_PKT_THROUGHPUT = 0x180
    BSL_422 = 0x40
    BSL_PAD_EN = 0x20
    BSL_ENDIAN_BIG = 0x0
    BSL_ENDIAN_LITTLE = 0x10
    BSL_REL_PKT_CNT = 0x8

    BSL_SYNC_COUNT_MASK = 0x7

    def __init__(self):
        NPDSegment.__init__(self)
        self.block_status = None
        self.sync_bytes = []
        self.data = b""

    def unpack(self, buffer):
        """
        Unpack a string buffer into an RS232 segment. Return the remaining buffer so that the next segment can iteratively
        be unpacked

        :param buffer: A string buffer representing an RS232 segment
        :type buffer: str

        :rtype: str
        """
        remaining = NPDSegment.unpack(self, buffer)
        (self.block_status,) = struct.unpack_from(">H", self.payload)
        sync_word_cnt = self.block_status & RS232Segment.BSL_SYNC_COUNT_MASK
        if sync_word_cnt > 0:
            self.sync_bytes = list(struct.unpack_from(">{}B".format(sync_word_cnt), self.payload[2:]))
            self.data = self.payload[2+sync_word_cnt:]
        else:
            self.data = self.payload[2:]
        return remaining

    def pack(self):
        """
        Pack the RS232 Segment object into a binary buffer

        :rtype: str
        """
        if self.block_status is None:
            raise Exception("block_status attribute should be defined")
        # Build the status word by calculating the sync bytes and adding to the rest of the header
        self.block_status = (self.block_status & 0xFFF8) + len(self.sync_bytes)
        self.payload = struct.pack(">H", self.block_status)
        # Add the sync words
        for sync_byte in self.sync_bytes:
            self.payload += struct.pack(">B", sync_byte)
        self.payload += self.data
        return NPDSegment.pack(self)

    def __eq__(self, other):
        if not isinstance(other, RS232Segment):
            return False
        for attr in ["timedelta", "segmentlen", "errorcode", "flags", "block_status", "sync_bytes", "data"]:
            if getattr(other, attr) != getattr(self, attr):
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "RS232 NPD Segment. TimeDelta={} Segment Len={} ErrorCode={:#0X} Flags={:#0X} Block_Status={:#0X} " \
               "DataLen={}" \
               "".format(self.timedelta, self.segmentlen, self.errorcode, self.flags, self.block_status, len(self.data))


class MIL1553Segment(NPDSegment):
    def __init__(self):
        NPDSegment.__init__(self)
        self.blockstatus = 0
        self.gap1 = 0
        self.gap2 = 0
        self.data = ""

    def unpack(self, buffer):
        remaining = NPDSegment.unpack(self, buffer)
        (self.blockstatus, self.gap2, self.gap2) = struct.unpack_from(">HBB", self.payload)
        self.data = self.payload[4:]
        return remaining

    def __repr__(self):
        return "MIL-STD-1553 Segment. TimeDelta={} Segment Len={} ErrorCode={:#0X} Flags={:#0X} BlockStatus={:#0X} " \
               "Gap1={} Gap2={}" \
               "".format(self.timedelta, self.segmentlen, self.errorcode, self.flags, self.blockstatus, self.gap1,
                         self.gap2)

class NPD(object):
    """ 
    Class to pack and unpack NPD payloads. 

    Capture a UDP packet and unpack the _payload as an NPD packet

    >>> import socket
    >>>> recv_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  
    >>> data, addr = recv_socket.recvfrom(2048)
    >>> n = NPD()
    >>> n.unpack(data)
    >>> print n.datatype
    6
    >>> for segment in n.segments:
    ...  print segment.errorcode
    2

    :type version: int
    :type hdrlen: int
    :type datatype: int
    :type packetlen: int
    :type cfgcnt: int
    :type flags: int
    :type sequence: int
    :type datasrcid: int
    :type mcastaddr: str
    :type timestamp: int
    :type segments: list[NPDSegment|ACQSegment|RS232Segment|A429Segment]
    """

    NPD_HEADER_FORMAT = '>BBHBBHIII'
    NPD_HEADER_LENGTH = struct.calcsize(NPD_HEADER_FORMAT)
    NPD_VERSION = 3

    NPD_DT = { 0x50: RS232Segment, 0x38: A429Segment, 0xA1: ACQSegment, 0xD0: MIL1553Segment}

    def __init__(self):
        '''Creator method for a UDP class'''
        self.version = NPD.NPD_VERSION #: Version
        self.hdrlen = NPD.NPD_HEADER_LENGTH//4  #: Header Length
        self.datatype = None  #: A unique identifier for the type of data collected in the packet
        self.packetlen = None  #: The number of 32-bit words in the data packet including the NPD header and data segments.
        self.cfgcnt = None  #: Stores an 8-bit number that is incremented (mod 256) each time the network device is configured.
        self.flags = None  #: Flags [0]-Unlocked timestamp [1]-Packet fragmentation [2]-Relative Time Count Present
        self.sequence = None  #: Sequence number
        self.datasrcid = None  #: A unique data source identifier for each data source.
        self.mcastaddr = ""  #: The 32-bit IP multicast address used as the destination address of the packet.
        self.timestamp = None  #: The content of this field is based upon the R bit in the flags field of the NPD Packet Protocol header.
        self.segments = []  #: List of all the data segments

    def unpack(self, buffer):
        """
        Unpack a string buffer into an NPD object
        
        :param buffer: A string buffer representing an NPD packet
        :type buffer: bytes
        :rtype: None
        """
        (_ver_hdr, self.datatype, self.packetlen, self.cfgcnt, self.flags, self.sequence, self.datasrcid,
         _mcast, self.timestamp) = struct.unpack_from(NPD.NPD_HEADER_FORMAT, buffer)
        self.version = _ver_hdr >> 4
        self.hdrlen = _ver_hdr & 0xF
        self.mcastaddr = inet_ntoa(struct.pack(">I",_mcast))

        _payload = buffer[self.hdrlen * 4:]

        if self.packetlen * 4 != len(buffer):
            raise Exception("The self reported packet length does not match the length of the buffer supplied")

        remain_buf = _payload
        while remain_buf != b"":
            if self.datatype in NPD.NPD_DT:
                segment = NPD.NPD_DT[self.datatype]()
            else:
                segment = NPDSegment()
            try:
                remain_buf = segment.unpack(remain_buf)
            except Exception as e:
                raise Exception(e)
            else:
                self.segments.append(segment)

        return True

    def pack(self):
        """
        Pack the NPD object into a binary buffer
        
        :rtype: str 
        """
        _ver_hdr = (self.version << 4) + self.hdrlen
        (_mc,) = struct.unpack(">I", inet_aton(self.mcastaddr))

        _payload = b""
        for segment in self.segments:
            _payload += segment.pack()
        self.packetlen = (NPD.NPD_HEADER_LENGTH + len(_payload))//4
        hdr_buf = struct.pack(NPD.NPD_HEADER_FORMAT, _ver_hdr, self.datatype, self.packetlen, self.cfgcnt, self.flags,
                              self.sequence, self.datasrcid, _mc, self.timestamp)

        return hdr_buf + _payload

    def __repr__(self):
        det = "NPD: DataType={:#0X} Seq={} DataSrcID={:#0X} MCastAddr={}".format(
            self.datatype, self.sequence, self.datasrcid, self.mcastaddr)
        for seg in self.segments:
            det += "\n\t{}".format(repr(seg))

        return det

    def __eq__(self, other):
        if not isinstance(other, NPD):
            return False
        for attr in ["version", "hdrlen", "datatype", "packetlen", "cfgcnt", "flags", "sequence", "datasrcid",
                     "mcastaddr", "timestamp", "segments"]:
            if getattr(other, attr) != getattr(self, attr):
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __iter__(self):
        self._index = 0
        return self

    def next(self):
        if self._index < len(self.segments):
            _dw = self.segments[self._index]
            self._index += 1
            return _dw
        else:
            raise StopIteration

    __next__ = next

    def __len__(self):
        return len(self.segments)

    def __getitem__(self, key):
        return self.segments[key]



#!/usr/bin/env python


# Very rudimentary (but fast) validation of recorded data
# Finds all inetx packets and validate no missing sequence numbers
# It will also take the url to an axn mem and verify data after downloading


import sys
sys.path.append("..")
import AcraNetwork.Pcap as pcap
import glob
import os.path
import struct
import time
import logging
import argparse
import json
from urllib.parse import urlparse
from urllib.request import urlopen
from os import mkdir, remove


VERSION = "0.1.4"

logging.basicConfig(level=logging.INFO, format='%(levelname)-6s %(asctime)-15s %(message)s')


def create_parser():
    # Argument parser
    parser = argparse.ArgumentParser(description='Validate inetx sequence numbers quickly in a pcap file')
    # Common
    parser.add_argument('--folder',  type=str, required=True, default=None, help='folder to parser for pcap files. Can be a URL')
    parser.add_argument('--verbose', required=False, action='store_true', default=False,    help="verbose mode")
    parser.add_argument('--control', type=int, required=False, default=0x11000000, help='control field value')
    parser.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION))

    return parser

# recorder data
# tcpdump -i eth0 -W 999 -C 1000 -n -K -s 2000 -w rec.pcap.


def uri_validator(x):
    try:
        result = urlparse(x)
        return all([result.scheme, result.netloc])
    except:
        return False


def main(args):
    roll_over = pow(2, 32)
    fnames = {}

    if uri_validator(args.folder):
        is_url = True
        all_files = []
        with urlopen(args.folder) as response:
            response_content = response.read()
            json_resp = json.loads(response_content)
            for f,e in json_resp.items():
                all_files.append(e["url"])
                fnames[e["url"]] = f
    else:
        is_url = False
        # Find all the files and sort by extension
        all_files = glob.glob(os.path.join(args.folder, "*.pcap*"))
    all_files.sort()

    # For recording the data
    stream_ids = {}
    reset_count = {}
    lost_sid_count = {}
    packet_lengths = {}
    inetx_pkts_validate = 0
    data_count_bytes = 0
    start_t = time.time()
    loss_count = 0

    for pfile in all_files:
        # To calculate the rate per pcap
        first_pcap_time = None
        last_pcap_time = None
        packet_data_vol = 0
        tmp_folder = "httpdl"
        loss = 0
        outf = ""

        if is_url:
            CHUNK = 32 * 1024
            if not os.path.exists(tmp_folder):
                mkdir(tmp_folder, 0o755)
            outf = os.path.join(tmp_folder , fnames[pfile])
            sd = time.time()
            with urlopen(pfile) as response, open(outf, 'wb') as out_file:
                data_len = 0
                while True:
                    chunk = response.read(CHUNK)
                    if not chunk:
                        break
                    data_len += len(chunk)
                    out_file.write(chunk)

                dlspeed = data_len * 8 / (time.time() - sd) / 1e6
                logging.info("Downloaded {} at {:.1f}Mbps and wrote to {}".format(pfile, dlspeed, outf))
            p = pcap.Pcap(outf)
        else:
            p = pcap.Pcap(pfile)
        for i, r in enumerate(p):
            if first_pcap_time is None:
                first_pcap_time = r.sec + r.usec * 1e-6
            last_pcap_time = r.sec + r.usec * 1e-6
            packet_data_vol += len(r.payload)
            if len(r.payload) >= (18+0x24):  # For short packets don't try to decode them as inetx
                # pull out the key fields
                (dst_port, udp_len, checksum, control, stream_id, seq) = struct.unpack_from(">HHHIII", r.payload, 0x24)
                if control == args.control:
                    if stream_id in stream_ids:
                        if seq != (stream_ids[stream_id] + 1) % roll_over:
                            if seq < stream_ids[stream_id]:
                                logging.warning("Source Restarted. File={} PktNum={} StreamID={:#0X} PrevSeq={} "
                                                "CurSeq={}".format(pfile, i, stream_id, stream_ids[stream_id], seq, ))
                                reset_count[stream_id] += 1
                            else:
                                loss = seq - ((stream_ids[stream_id] + 1) % roll_over)
                                logging.error("File={} PktNum={} StreamID={:#0X} PrevSeq={} CurSeq={} Lost={}".format(
                                    pfile, i, stream_id, stream_ids[stream_id], seq, loss))
                                loss_count += loss
                                lost_sid_count[stream_id] += loss
                    if stream_id not in reset_count:
                        reset_count[stream_id] = 0
                        lost_sid_count[stream_id] = 0
                        packet_lengths[stream_id]  = len(r.payload)
                    stream_ids[stream_id] = seq
                    inetx_pkts_validate += 1
                    data_count_bytes += len(r.payload)

        p.close()
        # The data rate at which we are validating
        dr = (data_count_bytes * 8)/(1e6 * (time.time() - start_t))
        try:
            ave_rec_rate_mbps = (packet_data_vol * 8) /(last_pcap_time - first_pcap_time) / 1e6
        except:
            ave_rec_rate_mbps = 0
        sids_found = len(stream_ids)
        info_str = "{} packets validated. Total_data={:.0f}MB Completed file {} Lost={} StreamsFound={} " \
                   "RecordRate={:.0f}Mbps ValRate={:.0f}Mbps".format(inetx_pkts_validate,  data_count_bytes/1e6, pfile, loss_count, sids_found,
                                       ave_rec_rate_mbps, dr)
        if loss > 0:
            logging.error(info_str)
        else:
            logging.info(info_str)
        if args.verbose:
            if len(stream_ids) > 0:
                logging.info("{:>7s} {:>9s} {:>9s} {:>9s} {:>9s}".format("SID", "Seq", "LostCount", "ResetCnt", "Length"))
            for s in sorted(stream_ids):
                logging.info("{:#07X} {:9d} {:9d} {:9d} {:9d}".format(s, stream_ids[s], lost_sid_count[s], reset_count[s], packet_lengths[s]))
        if is_url and loss == 0:
            remove(outf)
        elif is_url and not args.verbose:
            remove(outf)

    if len(stream_ids) > 0:
        logging.info("{:>7s} {:>9s} {:>9s} {:>9s} {:>9s}".format("SID", "Seq", "LostCount", "ResetCnt", "Length"))
    for s in sorted(stream_ids):
        logging.info("{:#07X} {:9d} {:9d} {:9d} {:9d}".format(s, stream_ids[s],  lost_sid_count[s], reset_count[s], packet_lengths[s]))

    print("\nSUMMARY: RXPKTS={} RXBYTES={} LOSTPKTS={}".format(inetx_pkts_validate, data_count_bytes, loss_count))


if __name__ == '__main__':
    parser = create_parser()
    args = parser.parse_args()
    main(args)
    sys.exit(0)



"""
.. module:: Chapter7
    :platform: Unix, Windows
    :synopsis: Class to construct and de construct Chapter7 Packets
    http://www.irig106.org/docs/106-17/chapter7.pdf

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""
__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import struct
from AcraNetwork import Golay
import math
import logging


ch7_logger = logging.getLogger(__name__)

PTDP_CONTENT_FILL = 0X0
PTDP_CONTENT_ASP = 0X1
PTDP_CONTENT_TEST_CNT = 0X2
PTDP_CONTENT_CH10 = 0X3
PTDP_CONTENT_MAC = 0X4
PTDP_CONTENT_IP = 0X5
PTDP_CONTENT_CH24 = 0X6

PTDP_CONTENT_TEXT = ["Fill", "ASP", "Test Counter", "Chapter 10", "Ethernet MAC", "IP", "Chapter 24"]

PTDP_FRAGMENT_COMPLETE = 0X0
PTDP_FRAGMENT_FIRST = 0X1
PTDP_FRAGMENT_MIDDLE = 0X2
PTDP_FRAGMENT_LAST = 0X3

PTDP_FRAGMENT_TEXT = ["Complete", "First", "Middle", "Last"]

PTDP_HDR_LEN = 0x6  # 24bits x2
PTFR_HDR_LEN = 0x4  # 1 byte unprotected and 3 bytes protected


PTDP_MAX_LEN = 0x800


class PTDPLengthError(Exception):
    pass


class PTDPRemainingData(Exception):
    pass


def _new_ptfr(ptfr_len, streamid=0x1, golay=Golay.Golay()):
    """Return a new PTFR object initialised with some useful values"""
    ptfr = PTFR(golay)
    ptfr.length = ptfr_len
    ptfr.streamid = streamid
    return ptfr


def datapkts_to_ptfr(eth_ch10_packets, ptfr_len=500, streamid=0x1, golay=Golay.Golay()):
    """
    Generator that will take a generator for ethernet packet aligned payloads
    and return the PTFRs encapsulating the payload

    :type eth_ch10_packets: collections.Iterable[bytes, bool]
    :param ptfr_len: Length of the PTFR frame
    :type ptfr_len: int
    :rtype: collections.Iterable[PTFR]
    """
    # Create a new frame
    ptfr = _new_ptfr(ptfr_len, streamid, golay)
    for ptdp in datapkts_to_ptdp(eth_ch10_packets):
        # Encapsulate the data in PTDP packets
        # Add the payload to the PTFR frame. IF we get a remainder then this from is full
        remainder = ptfr.add_payload(ptdp.pack(), ptdp.low_latency)
        while remainder != bytes():
            # Spit out the full frame
            yield ptfr
            # Create anew frame
            ptfr = _new_ptfr(ptfr_len, streamid, golay)
            # Maybe the  PTFP is bigger than one frame, then split it across PTFRs.
            while len(remainder) > ptfr_len:
                ptfr.ptdp_offset = 0x3ff  # Signifies that we have all PTDP payload
                ptfr.add_payload(remainder[:ptfr_len])
                yield ptfr  # Frame is full
                remainder = remainder[ptfr_len:]  # If we have more data then save it for the next frame
                ptfr = _new_ptfr(ptfr_len, streamid, golay)

            # Point to the first offset of the frame.
            ptfr.ptdp_offset = len(remainder)
            #Add the PTDS to the frame
            remainder = ptfr.add_payload(remainder)


def datapkts_to_ptdp(eth_ch10_packets):
    """
    Generator that will take a generator for ethernet packet aligned payloads
    and return the PTDPs encapsulating the payload

    :param eth_ch10_packets:
    :type eth_ch10_packets: collections.Iterable[bytes, bool]
    :rtype: collections.Iterable[PTDP]
    """
    for buffer, llp in eth_ch10_packets:
        # The packet fits in one PTDP
        if len(buffer) <= PTDP_MAX_LEN:
            ptdp_pkt = PTDP()
            ptdp_pkt.low_latency = llp
            ptdp_pkt.fragment = PTDP_FRAGMENT_COMPLETE
            ptdp_pkt.content = PTDP_CONTENT_MAC
            ptdp_pkt.payload = buffer
            ptdp_pkt.length = len(buffer)
            yield ptdp_pkt
        else:
            # NEED to split the packet into multiple PTDP packets. How many?
            number_of_packets = int(math.ceil(float(len(buffer)) / PTDP_MAX_LEN))
            for i in range(number_of_packets):
                ptdp_pkt = PTDP()
                ptdp_pkt.low_latency = llp
                ptdp_pkt.content = PTDP_CONTENT_MAC  # Assume it's Ethernet for the moment
                # Label the fragments
                if i == 0:
                    ptdp_pkt.fragment = PTDP_FRAGMENT_FIRST
                    ptdp_pkt.payload = buffer[:PTDP_MAX_LEN]
                elif i == number_of_packets - 1:
                    ptdp_pkt.fragment = PTDP_FRAGMENT_LAST
                    ptdp_pkt.payload = buffer[i*PTDP_MAX_LEN:]
                else:
                    ptdp_pkt.fragment = PTDP_FRAGMENT_MIDDLE
                    ptdp_pkt.payload = buffer[PTDP_MAX_LEN*i:PTDP_MAX_LEN*(i+1)]

                ptdp_pkt.length = len(ptdp_pkt.payload)
                yield ptdp_pkt


class PTDP(object):
    def __init__(self, golay=Golay.Golay()):
        self.payload = ""
        self.low_latency = False
        self.length = 0
        self.content = PTDP_CONTENT_FILL
        self.fragment = PTDP_FRAGMENT_COMPLETE
        self._payload = ""
        self._golay = golay

    @property
    def payload(self):
        return self._payload

    @payload.setter
    def payload(self, val):
        if len(val) > PTDP_MAX_LEN:
            raise Exception("One PTDP packet can only be of max length {}. Split your data into multiple PTDP packets".format(PTDP_MAX_LEN))

        self._payload = val

    def pack(self):
        """
        Convert a PTDP object into a string buffer

        :rtype: bytes
        """
        self.length = len(self.payload)
        msw = self.length & 0xFFF
        lsw = (self.length >> 12) + (self.fragment << 4) + (self.content << 6)

        return self._golay.encode(lsw, as_string=True) + self._golay.encode(msw, as_string=True) + self.payload

    def unpack(self, buffer):
        """
        Convert a buffer into a PTDP object returning the remaining buffer

        :type buffer: bytes
        :rtype: bytes
        """

        if len(buffer) < 6:
            raise PTDPRemainingData("Can't unpack less than the header length")

        lsw = self._golay.decode(buffer[:3])
        msw = self._golay.decode(buffer[3:6])

        self.length = msw + ((lsw & 0xF) << 12)
        self.fragment = (lsw >> 4) & 0x3
        self.content = (lsw >> 6) & 0xF
        if self.length > PTDP_MAX_LEN:
            raise PTDPLengthError("GolayHdr=len={}. Must be corrupted".format(self.length))
        elif len(buffer[6:]) < self.length:
            raise PTDPRemainingData("Buffer length={} GolayHdr=len={} fragment={} content={}".format(
                len(buffer[6:]), self.length, self.fragment, self.content))
        self.payload = buffer[6:self.length+6]

        return buffer[self.length+6:]

    def __len__(self):
        return len(self.payload) + PTDP_HDR_LEN

    def __eq__(self, other):
        if not isinstance(other , PTDP):
            return False
        for attr in ["payload", "length", "fragment", "content"]:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "PTDP: Len={} Content={} Fragment={} LowLatency={}".format(
            self.length, PTDP_CONTENT_TEXT[self.content], PTDP_FRAGMENT_TEXT[self.fragment], self.low_latency)


class PTFR(object):
    """
    Object to represent the PTFR frame
    Pass in a Golay object as creation of one is expensive so sharing and caching speeds things up a lot
    """
    def __init__(self, golay=Golay.Golay()):
        self.version = 0x0
        self.streamid = 0x0
        self.llp = False
        self.ptdp_offset = 0x0
        self.length = 0
        self._payload = bytes()
        self._golay = golay

    @property
    def payload(self):
        return self._payload

    @payload.setter
    def payload(self, val):
        if len(val) + len(self._payload) > self.length:
            raise Exception("Length of payload ({}) is larger than length field ({})".format(len(val), self.length))
        self._payload += val

    def add_payload(self, buffer, is_llp=False):
        """
        Add the buffer to the payload ensuring not to go over the length field
        :param buffer:
        :return:
        """
        if is_llp and len(self._payload) > 0 and self.llp:
            ch7_logger.debug("Adding an LLP buffer to some existing llp payload. Shifting original data and adding llp at "
                          "start. len={}".format(len(buffer) + 1))
            self.ptdp_offset += (len(buffer) + 1)
            self._payload = buffer + struct.pack(">B", 0xFF) + self._payload
            self.llp = True
        elif is_llp and len(self._payload) > 0 and not self.llp:
            ch7_logger.debug("Adding an LLP buffer to a payload with high latency data. Shifting the data and adding the "
                          "offset. len={}".format(len(buffer) + 1))
            self.ptdp_offset = (len(buffer) + 1)
            self._payload = buffer + struct.pack(">B", 0x0) + self._payload
            self.llp = True
        elif is_llp and len(self.payload) == 0:
            ch7_logger.debug("Adding first LLP buffer len={}".format(len(buffer)+1))
            self.llp = True
            self._payload = buffer + struct.pack(">B", 0xFF)
        else:
            ch7_logger.debug("Adding a normal PTDP buffer len={}".format(len(buffer)))
            self._payload += buffer

        if len(self._payload) > self.length:
            len_to_take = self.length-len(self._payload)
            remainder = self.payload[len_to_take:]
            self._payload = self.payload[:len_to_take]
            return remainder
        else:
            return bytes()

    def pack(self):
        """
        Convert a PTFR object into a string for transmission. This will return the packed string and the remainder string
        for any partial ptdp packet

        :param previous_partial: If we have some partial data from the previous PTFR packet, add it here
        :rtype: (bytes, bytes)
        """
        if len(self.payload) != self.length:
            raise Exception("Payload length does not match the length field")

        buffer = struct.pack(">B", self.version + (self.streamid << 4))
        buffer += self._golay.encode(self.ptdp_offset + (self.llp << 11), as_string=True)
        buffer += self.payload

        return buffer

    def unpack(self, buffer):
        """
        Convert the PTFR data from one minor frame into a PTFR object
        :param buffer:
        :return:
        """
        (byte_,) = struct.unpack_from(">B", buffer)
        self.version = byte_ & 0x3
        self.streamid = (byte_ >> 4) & 0xF
        # Protected field
        protected_field = self._golay.decode(buffer[1:4])
        self.llp = bool((protected_field >> 11) & 0x1)
        self.ptdp_offset = protected_field & 0x7FF
        self.payload = buffer[4:]

        return True

    def check_offsets(self, act_offset):
        if (act_offset != self.ptdp_offset) and (self.ptdp_offset != 2047):
            ch7_logger.error(
                "Offset of unpacked PTDP packet ({}) does not match the declared offset ({})".format(
                    act_offset, self.ptdp_offset))
        elif (self.ptdp_offset != 2047):
            ch7_logger.debug(
                "Offset of unpacked PTDP packet ({}) matches the declared offset ({})".format(
                    act_offset, self.ptdp_offset))

    def get_aligned_payload(self, first_PTFR, remainder=None):
        """
        Return the payload as PTDP packets with the final partial payload
        The remainder is the bytes from the end of the previous PTFR. IF this is the middle of a
        capture set it to None so that false positive messages about offsets is triggered

        :type remainder: bytes
        :param remainder: Optional partial payload from previous frame
        :rtype: Tuple[PTDP, bytes, str]
        """
        aligned = True
        # The PTFR decides what is low latency initially
        is_llp = self.llp

        if is_llp:
            ch7_logger.debug("LLP flag set. First packet should be LLP")
            buf = self.payload
        elif remainder is None and self.ptdp_offset > 0 and self.ptdp_offset < 0x7ff:
            buf = self.payload[self.ptdp_offset:]
            ch7_logger.debug("Start of analysis. Could be in the middle of a packet offset={} buffer length={}".format(self.ptdp_offset, len(buf)))
        elif remainder == bytes() and self.ptdp_offset > 0 and self.ptdp_offset < 0x7ff:
            buf = self.payload[self.ptdp_offset:]
            ch7_logger.debug("No remainder from previous packet, offset={} buffer length={}".format(self.ptdp_offset, len(buf)))
        elif remainder is None:
            buf =  self.payload
            ch7_logger.debug("Buffer length={}. Ignoring offset={} Remainder undefined".format(
                len(buf), self.ptdp_offset))
        else:
            buf = remainder + self.payload
            ch7_logger.debug("Buffer length={}. Ignoring offset={} Remainder length={}".format(
                len(buf), self.ptdp_offset, len(remainder)))

        if is_llp:
            byte_offset = 0
        elif remainder is None:  # Fake the offset if we have jumped into the middle of a data stream
            byte_offset = self.ptdp_offset
        else:
            byte_offset = -1 * len(remainder)
        do_offset_check = True
        offset_check_count = 0
        while aligned:
            p = PTDP(self._golay)
            try:
                buf = p.unpack(buf)
            except PTDPLengthError as e:
                aligned = False
                ch7_logger.warning("Looks like we got an illegal PTDP length. Resetting. {}bytes. Message={} Offset={}".format(len(buf), e,
                                                                                                       self.ptdp_offset))
                yield (None, None, e)
            except PTDPRemainingData as e:
                aligned = False
                ch7_logger.debug("Failed to unpack buffer of length {}bytes. Message={} Offset={}".format(len(buf), e,
                                                                                                       self.ptdp_offset))
                if not is_llp and byte_offset > 0 and do_offset_check:
                    self.check_offsets(byte_offset)

                yield (None, buf, e)
            else:
                if not is_llp and do_offset_check and byte_offset >= 0:
                    self.check_offsets(byte_offset)
                    do_offset_check = False
                    offset_check_count += 1
                elif not is_llp and not do_offset_check and offset_check_count < 1:
                    do_offset_check = True
                    byte_offset += len(p)
                elif not is_llp:
                    byte_offset += len(p)


                # set the low latency flag on the current packet now we know if we are at the end of the LLP sequence.
                p.low_latency = is_llp


                if is_llp:  # If this is a low latency packet
                    # Remove the last byte
                    (next_llp, ) = struct.unpack_from(">B", buf)
                    # Check if the next PTDP is low latency before yielding
                    if next_llp == 0xFF:
                        is_llp = True
                        buf = buf[1:]
                        byte_offset += (len(p) + 1)
                    else:
                        is_llp = False
                        #if ((remainder == bytes()) or first_PTFR)  and self.ptdp_offset > 0:
                        if ( ((remainder == bytes()) and self.ptdp_offset > 0) or first_PTFR):
                            ch7_logger.debug( "LLP Packets extracted, jumping to offset")
                            buf = self.payload[self.ptdp_offset:]
                            do_offset_check = False
                            byte_offset = self.ptdp_offset
                            offset_check_count = 1
                        elif remainder is None:
                            buf =  buf[1:]
                            byte_offset += (len(p) + 1)
                        else:
                            buf = remainder + buf[1:]  # The remainder is only added after all the llp packets are removed
                            byte_offset += (len(p) + 1 - len(remainder))
                            if len(remainder) > 0:
                                do_offset_check = False

                yield (p, bytes(), "")

    def __eq__(self, other):
        if not isinstance(other, PTFR):
            return False

        for attr in ["version", "streamid", "llp", "ptdp_offset", "payload"]:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "PTFR: Length={} StreamID={:#0X} Offset={} LowLatency={}\n".format(
            self.length, self.streamid, self.ptdp_offset, self.llp)



"""
.. module:: Golay
    :platform: Unix, Windows
    :synopsis: Encode and decode Golay Codes based on IRIG106 Standard http://www.irig106.org/docs/106-17/chapter7.pdf

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""
__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2020"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


# This is a direct porting of the C code from the IRIG106 starndard.
import struct
from functools import lru_cache

GOLAY_SIZE = 0x1000

G_P = [
0xc75, 0x63b, 0xf68, 0x7b4,
0x3da, 0xd99, 0x6cd, 0x367,
0xdc6, 0xa97, 0x93e, 0x8eb]

H_P = [
0xa4f, 0xf68, 0x7b4, 0x3da,
0x1ed, 0xab9, 0xf13, 0xdc6,
0x6e3, 0x93e, 0x49f, 0xc75
]


class Golay():
    """
    Encode and Decode Golay numbers
    """
    def __init__(self):
        self.SyndromeTable = [0] * GOLAY_SIZE
        self.CorrectTable = [0] * GOLAY_SIZE
        self.ErrorTable = [0] * GOLAY_SIZE

    
    @staticmethod
    @lru_cache()
    def _init_Table():
        EncodeTable = [0] * GOLAY_SIZE
        for x in range(GOLAY_SIZE):
            EncodeTable[x] = (x << 12)
            for i in range(12):
                if (x >> (11 - i) & 1):
                    EncodeTable[x] ^= G_P[i]

        return EncodeTable

    @lru_cache(maxsize=20)
    def encode(self, raw, as_string=False):
        """
        Encode the value as a 24b code

        :type raw: int
        :return: int
        """
        if 0xFFF < raw < 0:
            raise Exception("Converestion of 12b value only")

        EncodeTable = Golay._init_Table()
        encoded = EncodeTable[raw & 0xfff]
        if as_string:
            return struct.pack(">BH", encoded >> 16, encoded & 0xFFFF)
        else:
            return encoded
            
    @lru_cache(maxsize=20)
    def decode(self, encoded):
        """
        Decode a 24b number as a golay

        :type encoded: int|bytes
        :param encoded:
        :return:
        """
        if type(encoded) is bytes:
            if len(encoded) != 3:
                raise Exception("String to decode should be 3 bytes")
            (b,w) = struct.unpack(">BH", encoded)
            v = w + (b << 16)
        elif 0xFFFFF < encoded < 0:
            raise Exception("Only supports 24b unsigned numbers")
        else:
            v = encoded

        self._initgolaydecode()
        return self._decode2(((v) >> 12) & 0xfff, (v) & 0xfff)

    def _syndrome2(self, v1, v2):
        return self.SyndromeTable[v2] ^ (v1)

    def _syndrome(self, v):
        return self._syndrome2(((v) >> 12) & 0xfff, (v) & 0xfff)

    def _errors2(self, v1, v2):
        return self.ErrorTable[self._syndrome2(v1, v2)]

    def _decode2(self, v1, v2):
        return (v1) ^ self.CorrectTable[self._syndrome2(v1, v2)]

    def _errors(self, v):
        return self._errors2(((v) >> 12) & 0xfff, (v) & 0xfff)

    @staticmethod
    def _onesincode(code, size):
        """Optimised version of the code below. Runs 2x"""
        return bin(code)[2:size+2].count('1')

    @staticmethod
    def _onesincode_old(code, size):
        ret = 0

        for t in range(size):
            if (code >> t) & 1:
                ret += 1

        return ret

    @lru_cache()
    def _initgolaydecode(self):
        for x in range(GOLAY_SIZE):
            self.SyndromeTable[x] = 0
            for i in range(12):
                if ( x >> (11-i)) & 1:
                    self.SyndromeTable[x] ^= H_P[i]
                    self.ErrorTable[x] = 4
                    self.CorrectTable[x] = 0xFFF

        self.ErrorTable[0] = 0
        self.CorrectTable[0] = 0
        for i in range(24):
            for j in range(24):
                for k in range(24):
                    error = (1 << i) | (1 << j) | (1 << k)
                    syndrom = self._syndrome(error)
                    self.CorrectTable[syndrom] = (error >> 12) & 0xfff
                    self.ErrorTable[syndrom] = Golay._onesincode(error, 24)

        return True



"""
.. module:: Chapter10
    :platform: Unix, Windows
    :synopsis: Class to construct and de construct Chapter10 Packets

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""
__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import struct
from functools import reduce
from datetime import datetime
import time


def get_checksum_buf(buf):
    """
    Return the arithmetic checksum of a header

    :param buf:
    :return:
    """
    if len(buf) % 2 != 0:
        raise Exception("buffer needs to be 16-bit aligned")

    words = struct.unpack("<{}H".format(len(buf)//2), buf)
    sum = reduce(lambda x, y: x + y, words)

    return sum % 65536


def double_digits_to_bcd(val):
    """
    Very simplified conversion of time to format used in Ch10 Time packet

    :param val: time field value in integers.
    :return:
    """
    offsets = {1: 0, 10: 4}
    retval = 0
    for dec, offset in offsets.items():
        retval += ((int(val / dec) % 10) << offset)
    return retval


def bcd_to_int(val):
    """

    :param val:
    :return:
    """
    if val < 0:
        raise ValueError("Cannot be a negative integer")

    if val == 0:
        return 0

    bcdstring = ''
    while val > 0:
        nibble = val % 16
        bcdstring = str(nibble) + bcdstring
        val >>= 4
    return int(bcdstring)


def buf_to_printable(buffer):
    pw = ""
    for idx in range(len(buffer)):
        (v,) = struct.unpack_from(">B", buffer, idx)
        pw += "{:#04X} ".format(v)
        if idx % 8 == 7:
            pw += "\n"
    return pw


class Chapter10UDP(object):
    """ 
    Class to encapsulate Chapter10 payload in UDP packets

    Capture a UDP packet and unpack the payload as an Chapter 10 packet
    
    There are two types of packets, segmented and full. 

    >>> import socket
    >>>> recv_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  
    >>> data, addr = recv_socket.recvfrom(2048)
    >>> n = Chapter10UDP()
    >>> n.unpack(data)
    >>> print n.type
    0

    :type version: int
    :type type: int
    :type hdrlen: int
    :type channelID: int
    :type channelsequence: int
    :type segmentoffset: int
    :type chapter10: Chapter10
    """

    CH10_UDP_HEADER_FORMAT1 = '<BBH'
    CH10_UDP_HEADER_FORMAT2 = ">HBB"
    CH10_UDP_SEG_HEADER_FORMAT1 = '<HBBI'

    CH10_HDR_LEN = {1: 4, 2: 12, 3: 8}  # fmt1, 2, 3

    CH10_UDP_HEADER_LENGTH = struct.calcsize(CH10_UDP_HEADER_FORMAT1)
    CH10_UDP_SEG_HEADER_LENGTH = struct.calcsize(CH10_UDP_SEG_HEADER_FORMAT1)

    TYPE_FULL = 0  #: Full Chapter 10 packets type field constant. Assign to :attr:`Chapter10UDP.type`
    TYPE_SEG = 1  #: Segmented Chapter 10 packets type field constant. Assign to :attr:`Chapter10UDP.type`

    def __init__(self):
        '''Creator method for a UDP class'''
        self.version = 1  #: Version
        self.type = None  #: Type of message , Full or Segmented
        self.channelID = None  #: Segmented Packets Only. Channel ID of the data in the RCC 106 Chapter 10 packet
        self.channelsequence = None  #: Segmented Packets Only, Channel Sequence Number of the data in the RCC 106 Chapter 10 packet
        self.sequence = None  #: UDP Sequence number
        self.segmentoffset = None  #: Segmented Packets Only. The 32-bit Segmented Packets Only, Position of the data in the RCC 106 Chapter 10 packet.
        self.packetsize = None  #: Format 2 Packet size
        self.sourceid_len = None  #: Format 3 Source ID length
        self.sourceid = 0  #: Format 3 Source ID
        self.offset_pkt_start = None  #: Format 3 Offset to packet start in bytes
        self.payload = b""
        self.chapter10 = Chapter10() #: The encapsulated Chapter10 packet. :class:`Chapter10`

    def unpack(self, buffer):
        """
        Unpack a string buffer into an Chapter10UDP object

        :param buffer: A string buffer representing an Chapter10UDP packet
        :type buffer: bytes
        :rtype: None
        """
        # Format 2 is big endian so check if this packet is format 2 first
        (seg_upr, seg_lwr, _ver_type) = struct.unpack_from(Chapter10UDP.CH10_UDP_HEADER_FORMAT2, buffer)
        # Because of the endianness issue, some format 1 packets can look like format 2 if we naievly look at the
        # format field. Instead also look at the size and if it matches the size of the buffer then assume that
        # we have the correct format
        (_size_upp, _size_lower,) = struct.unpack_from(">BH", buffer, 5)
        size_guess = _size_lower + (_size_upp << 16)
        if _ver_type & 0xF == 2 and size_guess == len(buffer)/4 - 3:
            self.version = _ver_type & 0xF
        else:
            (_ver_type, seg_lwr, seg_upr) = struct.unpack_from(Chapter10UDP.CH10_UDP_HEADER_FORMAT1, buffer)
            self.version = _ver_type & 0xF

        self.type = _ver_type >> 4
        self.sequence = seg_lwr + (seg_upr << 8)

        if self.version == 3:
            self.sourceid_len =_ver_type >> 4
            self.offset_pkt_start = seg_upr

        if self.type == Chapter10UDP.TYPE_SEG and self.format == 1:
            (self.channelID, self.channelsequence, _res, self.segmentoffset) = \
                struct.unpack_from(Chapter10UDP.CH10_UDP_HEADER_FORMAT1, buffer, Chapter10UDP.CH10_UDP_HEADER_LENGTH)
            self.payload = buffer[(Chapter10UDP.CH10_UDP_HEADER_LENGTH + Chapter10UDP.CH10_UDP_SEG_HEADER_LENGTH):]
        elif self.format == 1:
            self.payload = buffer[Chapter10UDP.CH10_UDP_HEADER_LENGTH:]
        elif self.format == 2:

            (_segoffset_upper ,_size_upp, _size_lower , _segoff_lower, self.channelID) = struct.unpack_from(
                ">BBHHH", buffer, 4
            )
            self.packetsize = _size_lower + (_size_upp << 16)
            self.segmentoffset = _segoff_lower + (_segoffset_upper << 16)
            self.payload = buffer[Chapter10UDP.CH10_UDP_HEADER_LENGTH + 8:]
        elif self.format == 3:
            (_srcid_datat, ) = struct.unpack_from("<I", buffer, 4)
            if self.sourceid_len == 0:
                self.sourceid = 0x0
                self.sequence = _srcid_datat
            elif self.sourceid_len == 1:
                self.sourceid = _srcid_datat >> (32 - 4)
                self.sequence = _srcid_datat & 0x0FFFFFFF
            elif self.sourceid_len == 2:
                self.sourceid = _srcid_datat >> (32 - 8)
                self.sequence = _srcid_datat & 0x00FFFFFF
            elif self.sourceid_len == 3:
                self.sourceid = _srcid_datat >> (32 - 12)
                self.sequence = _srcid_datat & 0x000FFFFF
            elif self.sourceid_len == 4:
                self.sourceid = _srcid_datat >> (32 - 16)
                self.sequence = _srcid_datat & 0x0000FFFF
            else:
                raise Exception("Source id length {} is not valid".format(self.sourceid_len))

            self.payload = buffer[Chapter10UDP.CH10_UDP_HEADER_LENGTH + 4:]

        else:
            self.payload = buffer[Chapter10UDP.CH10_UDP_HEADER_LENGTH:]

        return self.chapter10.unpack(self.payload)

    @property
    def format(self):
        return self.version

    @format.setter
    def format(self, val):
        self.version = val

    def pack(self):
        """
        Pack the Chapter10UDP object into a binary buffer

        :rtype: bytes
        """

        if self.format == 3:
            _ver_type = (self.sourceid_len << 4) + self.version
            seg_up = self.offset_pkt_start
            seg_lr = 0
        else:
            _ver_type = (self.type << 4) + self.version
            seg_up = self.sequence >> 8
            seg_lr = self.sequence & 0xFF

        if self.format == 2:
            _payload = struct.pack(Chapter10UDP.CH10_UDP_HEADER_FORMAT2, seg_up, seg_lr, _ver_type)
        else:
            _payload= struct.pack(Chapter10UDP.CH10_UDP_HEADER_FORMAT1, _ver_type, seg_lr, seg_up)

        self.payload = self.chapter10.pack()
        if self.type == Chapter10UDP.TYPE_SEG and self.format == 1:
            _payload += struct.pack(Chapter10UDP.CH10_UDP_SEG_HEADER_FORMAT1, self.channelID, self.channelsequence, 0, self.segmentoffset)

        elif self.format == 2:
            self.packetsize = len(self.payload)//4
            _payload += struct.pack(">BBHHH", self.segmentoffset >> 16, self.packetsize >> 16, self.packetsize & 0xFFFF, self.segmentoffset & 0xFFFF,
                                   self.channelID)

        elif self.format == 3:
            if self.sourceid_len == 0:
                _field = self.sequence
            elif self.sourceid_len == 1:
                _field = (self.sequence & 0x0FFFFFFF) + (self.sourceid << (32 - 4))
            elif self.sourceid_len == 2:
                _field = (self.sequence & 0x00FFFFFF) + (self.sourceid << (32 - 8))
            elif self.sourceid_len == 3:
                _field = (self.sequence & 0x000FFFFF) + (self.sourceid << (32 - 12))
            elif self.sourceid_len == 4:
                _field = (self.sequence & 0x0000FFFF) + (self.sourceid << (32 - 16))
            else:
                _field = 0
                raise Exception("Invalid source id")

            _payload += struct.pack("<I", _field)

        return _payload + self.payload

    def __repr__(self):
        if self.type == Chapter10UDP.TYPE_FULL:
            return "CH10 UDP Full Packet: Format={} Sequence={} Payload={}".format(
                self.format, self.sequence, repr(self.chapter10))
        else:
            return "CH10 UDP Sequence: Format={} Sequence={} ChID={} ChSeqNum={} SegOffset={}".format(
                self.format, self.sequence, self.channelID, self.channelsequence, self.segmentoffset)

    def __eq__(self, other):
        if not isinstance(other, Chapter10UDP):
            return False

        if self.type == Chapter10UDP.TYPE_SEG and self.format == 1:
            _match_att = ("version", "type", "sequence", "channelID", "channelsequence", "segmentoffset", "payload")
        elif self.format == 2:
            _match_att = ("format", "type", "sequence", "channelID", "channelsequence", "segmentoffset", "packetsize",
                          "payload")
        elif self.format == 3:
            _match_att = ("format", "sourceid_len", "sourceid", "sequence", "offset_pkt_start", "payload")
        else:
            _match_att = ("version", "type", "sequence", "payload")

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True


DATA_TYPE_TIMEFMT_1 = 0X11
DATA_TYPE_TIMEFMT_2 = 0X12
DATA_TYPE_PCM_DATA_FMT1 = 0X9


class Chapter10(object):
    """
    Class to pack and unpack Chapter10 payloads. 
    
    Create a packet and transmit it via UDP
    
    >>> import socket
    >>> # Open a socket
    >>> tx_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    >>> # Create the Ch10 UDP wrapper
    >>> ch10_udp = Chapter10UDP()
    >>> ch10_udp.type = Chapter10UDP.TYPE_FULL
    >>> ch10_udp.sequence = 1
    >>> # Populate the Chapter 10 packet inthe wrapper
    >>> ch10_udp.chapter10.channelID = 1
    >>> ch10_udp.chapter10.datatypeversion = 2
    >>> ch10_udp.chapter10.sequence = 3
    >>> ch10_udp.chapter10.packetflag = 0 # No secondary
    >>> ch10_udp.chapter10.datatype = 4
    >>> ch10_udp.chapter10.relativetimecounter = 100
    >>> ch10_udp.chapter10.payload = struct.pack(">II", 33, 44)
    >>> # Send the packet
    >>> tx_socket.sendto(ch10_udp.pack(), ("127.0.0.1", 8010))
    36

    :type syncpattern: int
    :type channelID: int
    :type packetlen: int
    :type datalen: int
    :type datatypeversion: int
    :type sequence: int
    :type datatype: int
    :type relativetimecounter: int
    :type timestamp: int
    :type ts_source: str
    :type payload: bytes
    :type data_checksum_size: int
    :type filler: bytes
    """

    SYNC_WORD = 0xEB25  #:(Object Constant) Sync word

    CH10_HDR_FORMAT = "<HHIIBBBBIHH"
    CH10_HDR_FORMAT_LEN = struct.calcsize(CH10_HDR_FORMAT)

    CH10_OPT_HDR_FORMAT = "<IIHH"
    CH10_OPT_HDR_FORMAT_LEN = struct.calcsize(CH10_OPT_HDR_FORMAT)

    TS_SOURCES = ["rtc", "ch4", "ieee1588"] #:(Object Constant) Valid timesources, assign to :attr:`Chapter10.ts_source`

    PKT_FLAG_SECONDARY = 0x80  #:(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_SEC_HDR_TIME = 0x40  #:(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_RTC_SYNC_ERROR = 0x20  #:(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_DATA_OVERFLOW = 0X10  #:(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_RCC_TIME = 0X0  #:(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_1588_TIME = 0X04  #:(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_8BIT_CHKSUM = 0X1  #:(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_16BIT_CHKSUM = 0X2  #:(Object Constant) add to :attr:`Chapter10.packetflag` to enable
    PKT_FLAG_32BIT_CHKSUM = 0X3  #:(Object Constant) add to :attr:`Chapter10.packetflag` to enable

    def __init__(self):

        self.syncpattern = Chapter10.SYNC_WORD  #:(2 Bytes) contains a static sync value for the every packet. The Packet Sync Pattern value shall be 0xEB25
        self.channelID = None  #:(2 Bytes) contains a value representing the Packet Channel ID.
        self.packetlen = None  #:(4 Bytes) contains a value representing the length of the entire packet. The value shall be in bytes and is always a multiple of four
        self.datalen = None  #:(4 Bytes) contains a value representing the valid data length within the packet
        self.datatypeversion = 0x5  #: RCC released versions
        self.sequence= None  #:(1 Byte) contains a value representing the packet sequence number for each Channel ID.
        self._packetflag = None  #:(1 Byte) contains bits representing information on the content and format of the packet(s)
        self.datatype = None  #:(1 Byte) contains a value representing the type and format of the data
        self.relativetimecounter = None  #:(6 Bytes) contains a value representing the 10 MHz Relative Time Counter (RTC)
        self.ptptimeseconds = None  #: PTP Timestamp seconds
        self.ptptimenanoseconds = None  #: PTP Timestamp nanoseconds
        self.ts_source = "ieee1588" #:The timestamp source. Select from :attr:`Chapter10.TS_SOURCES`
        self.payload = b""  #:The payload
        self.data_checksum_size = 0
        self.filler = b""
        self._secondary_header = False

    @property
    def packetflag(self):
        """
        (1 Byte) contains bits representing information on the content and format of the packet(s)

        :type val: intx
        """
        return self._packetflag

    @packetflag.setter
    def packetflag(self, val):

        if val > 0xFF:
            raise Exception("Packet flag ={:#0X} to valuid".format(val))
        self._packetflag = val

        if self._packetflag >> 7 == 1:

            if ((self._packetflag >> 2) & 0x3) == 0:
                self.ts_source = "ch4"
            elif ((self._packetflag >> 2) & 0x3) == 1:
                self.ts_source = "ieee1588"
            else:
                raise Exception("Time format is illegal")

            self._secondary_header = True
        else:
            self._secondary_header = False

    def pack(self):
        """
        Pack the Chapter10 object into a binary buffer

        :rtype: bytes
        """

        if self._secondary_header:
            if self.ts_source == "ch4":
                raise Exception("Ch4 Timestamp in secondary header not supported")

            sec_hdr = struct.pack(Chapter10.CH10_OPT_HDR_FORMAT, self.ptptimenanoseconds, self.ptptimeseconds, 0, 0)
            # Replace the checksum
            cs = get_checksum_buf(sec_hdr)
            sec_hdr = sec_hdr[:-2] + struct.pack("<H", cs)
        else:
            sec_hdr = b""

        total_len_excl_filler = len(sec_hdr) + len(self.payload) + Chapter10.CH10_HDR_FORMAT_LEN + self.data_checksum_size

        # Add the filler
        if total_len_excl_filler % 4 == 0:
            self.filler = b""
        else:
            fill_len = 4 - (total_len_excl_filler % 4)
            fill_val = [0xFF] * fill_len
            self.filler = struct.pack(">{}B".format(fill_len), *fill_val)

        self.packetlen = total_len_excl_filler + len(self.filler)
        self.datalen = len(self.payload)

        checksum = 0
        _rtc_lwr = self.relativetimecounter & 0xFFFFFFFF
        _rtc_upr = self.relativetimecounter >> 32

        hdr = struct.pack(Chapter10.CH10_HDR_FORMAT, self.syncpattern, self.channelID, self.packetlen, self.datalen,
                          self.datatypeversion, self.sequence, self.packetflag, self.datatype, _rtc_lwr,_rtc_upr, checksum)
        hdr = hdr[:-2] + struct.pack("<H", get_checksum_buf(hdr))
        return hdr + sec_hdr + self.payload + self.filler

    def unpack(self, buffer):
        """
        Unpack a string buffer into an Chapter10 object

        :param buffer: A string buffer representing an Chapter10 packet
        :type buffer: bytes
        :rtype: None
        """
        (self.syncpattern, self.channelID, self.packetlen, self.datalen, self.datatypeversion, self.sequence,
         self.packetflag, self.datatype, _rtc_lwr,_rtc_upr, checksum) = struct.unpack_from(Chapter10.CH10_HDR_FORMAT, buffer)

        offset_hdr = struct.calcsize(Chapter10.CH10_HDR_FORMAT)-2
        exp_checksum = get_checksum_buf(buffer[:offset_hdr])
        if checksum != exp_checksum:
            raise Exception("Ch10 Header checksum {:#0X} does not match expected={:#0X}".format(checksum, exp_checksum))
            #print("Ch10 Header checksum does not match expected={:#0X}".format(exp_checksum))

        self.relativetimecounter = _rtc_lwr + (_rtc_upr << 32)

        self.ts_source = "ieee1588"
        if (self.packetflag >> 7) == 1:
            self._secondary_header = True
            pkt_hdr_time =  (self.packetflag >> 2) & 0x3
            (ts_ns, ts_s, _res, _checksum_sec) = struct.unpack_from(
                Chapter10.CH10_OPT_HDR_FORMAT, buffer, Chapter10.CH10_HDR_FORMAT_LEN)

            sec_exp_checksum = get_checksum_buf(
                buffer[Chapter10.CH10_HDR_FORMAT_LEN:Chapter10.CH10_HDR_FORMAT_LEN+Chapter10.CH10_OPT_HDR_FORMAT_LEN-2])

            if _checksum_sec != sec_exp_checksum:
                raise Exception("Ch10 Secondary Header checksum ({:#0X}) does not match expected={:#0X}".format(
                    _checksum_sec, sec_exp_checksum))
                #print("Ch10 Secondary Header checksum does not match expected={:#0X}".format(sec_exp_checksum))

            if pkt_hdr_time == 0:
                raise Exception("Ch4 Timestamp in secondary header not supported")
                #print("Ch4 Timestamp in secondary header not supported")
            elif pkt_hdr_time == 1:
                self.ts_source = "ieee1588"
                self.ptptimenanoseconds = ts_ns
                self.ptptimeseconds = ts_s
            else:
                raise Exception("Secondary Header Time Format not legal")
                #print("Secondary Header Time Format not legal")
            self.payload = buffer[(Chapter10.CH10_HDR_FORMAT_LEN + Chapter10.CH10_OPT_HDR_FORMAT_LEN):]
        else:
            self._secondary_header = False
            self.payload = buffer[Chapter10.CH10_HDR_FORMAT_LEN:]


        return True

    def __eq__(self, other):
        if not isinstance(other, Chapter10):
            return False

        _match_att = ("syncpattern", "channelID", "packetlen", "datalen", "datatypeversion", "datatype", "_packetflag",
                      "relativetimecounter", "ptptimeseconds", "ptptimenanoseconds", "ts_source", "payload",
                      "data_checksum_size", "filler", "_secondary_header")

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        return "Chapter 10: ChannelID={} Sequence={} DataLen={}".format(self.channelID, self.sequence, self.datalen)


ITS_FREEWHEELING = (0X0 << 12)
ITS_FREEWHEELING_FROM_TIME = (0X0 << 12)
ITS_FREEWHEELING_FROM_RMM = (0X0 << 12)
ITS_LOCKED_IRIG = (0X0 << 12)
ITS_LOCKED_GPS = (0X0 << 12)
ITS_LOCKED_NTP = (0X0 << 12)
ITS_LOCKED_PTP = (0X0 << 12)
ITS_LOCKED_EMBEDDED = (0X0 << 12)

DATE_FMT_YEAR_AVAIL = (0X1 << 9)
DATE_FMT_LEAP_YEAR = (0X1 << 8)

TIME_FMT_IRIGB = 0X0
TIME_FMT_IRIGA = (0X1 << 4)
TIME_FMT_IRIGG = (0X2 << 4)
TIME_FMT_RTC = (0X3 << 4)
TIME_FMT_UTC = (0X4 << 4)
TIME_FMT_GPS = (0X5 << 4)
TIME_FMT_NONE = (0XF << 4)

SRC_INTERNAL = 0X0
SRC_EXTERNAL = 0X1
SRC_INT_FMM = 0X2
SRC_NONE = 0XF


class TimeDataFormat1(object):
    """
    Class to pack and unpack Chapter10 TIME packet payloads. 
    
    Create a packet and transmit it via UDP


    :type channel_specific_data: int
    :type milliseconds: int
    :type datetime: datetime | None
    :type filler: str
    """

    def __init__(self):

        self.channel_specific_data = ITS_FREEWHEELING + DATE_FMT_YEAR_AVAIL + TIME_FMT_GPS + SRC_EXTERNAL
        self.datetime = None
        self.milliseconds = 0

    def pack(self):
        """
        Pack the Chapter10 object into a binary buffer

        :rtype: str
        """
        if not isinstance(self.datetime, datetime):
            raise Exception("datetime attribute should be an instance of the datetime object ")

        packet_bytes = list()
        packet_bytes.append(double_digits_to_bcd(self.milliseconds/10))
        packet_bytes.append(double_digits_to_bcd(self.datetime.second))
        packet_bytes.append(double_digits_to_bcd(self.datetime.minute))
        packet_bytes.append(double_digits_to_bcd(self.datetime.hour))
        if (self.channel_specific_data & DATE_FMT_YEAR_AVAIL) >> 9 == 0x1:
            packet_bytes.append(double_digits_to_bcd(self.datetime.day))
            packet_bytes.append(double_digits_to_bcd(self.datetime.month))
            packet_bytes.append(double_digits_to_bcd(self.datetime.year % 100))
            packet_bytes.append(double_digits_to_bcd(self.datetime.year / 100))
        else:
            doy = int(self.datetime.strftime("%j"))
            packet_bytes.append(double_digits_to_bcd(doy % 100))
            packet_bytes.append(double_digits_to_bcd(doy / 100))

        mybytes = struct.pack("<I", self.channel_specific_data)
        mybytes += struct.pack("<{}B".format(len(packet_bytes)), *packet_bytes)

        return mybytes

    def unpack(self, buf):
        """
        Unpack a string buffer into a TimeFormat object
        :param buffer:
        :return:
        """
        (self.channel_specific_data, ms, s, mn, h) = struct.unpack_from("<IBBBB", buf)
        self.milliseconds = bcd_to_int(ms) * 10
        sec = bcd_to_int(s)
        mins = bcd_to_int(mn)
        hrs = bcd_to_int(h)

        if (self.channel_specific_data & DATE_FMT_YEAR_AVAIL) >> 9 == 0x1:
            (d, mo, yr) = struct.unpack_from("<BBH", buf, 8)
            day = bcd_to_int(d)
            mon = bcd_to_int(mo)
            year = bcd_to_int(yr)
            self.datetime = datetime(year, mon, day, hrs, mins, sec)
        else:
            (doy,hdoy) = struct.unpack_from("<BB", buf, 8)
            day_of_year = bcd_to_int(doy) + 100 * bcd_to_int(hdoy)
            date_as_string = "{:02d}:{:02d}:{:02d} {:03d} 1970".format(hrs, mins, sec, day_of_year)
            self.datetime = datetime.strptime(date_as_string, "%H:%M:%S %j %Y")

        return True

    def __eq__(self, other):
        if not isinstance(other, TimeDataFormat1):
            return False

        _match_att = ("channel_specific_data", "datetime", "milliseconds")

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        return "TimeFormat1 ChannelSpecificWord={:#0X} Time={} MilliSeconds={}".format(
            self.channel_specific_data, self.datetime.strftime("%H:%M:%S %x %d-%b %Y"), self.milliseconds
        )

    def __len__(self):
        if (self.channel_specific_data & DATE_FMT_YEAR_AVAIL) >> 9 == 0x1:
            return 4 * (4 + 1)
        else:
            return 4 * (3 + 1)


TS_STATUS_VALID = 0x1
TS_STATUS_IEEE2002 = (0X1 << 4)
TS_STATUS_IEEE2008 = (0X2 << 4)


class TimeDataFormat2(object):
    """
    Class to pack and unpack Chapter10 TIME packet payloads.

    Create a packet and transmit it via UDP


    :type channel_specific_data: int
    :type milliseconds: int
    :type datetime: datetime | None
    :type filler: str
    """

    def __init__(self):

        self.channel_specific_data = TS_STATUS_VALID + TS_STATUS_IEEE2002
        self.datetime = None
        self.nanoseconds = 0

    def pack(self):
        """
        Pack the Chapter10 time object into a binary buffer

        :rtype: str
        """
        if not isinstance(self.datetime, datetime):
            raise Exception("datetime attribute should be an instance of the datetime object ")
        utc_seconds = int(time.mktime(self.datetime.timetuple()))
        if (self.channel_specific_data >> 4) & 0x1:
            # ptp
            frac_sec = self.nanoseconds
        else:
            #ntp
            frac_sec = int(self.nanoseconds * (pow(2, 32)/1e9))

        return struct.pack("<III", self.channel_specific_data, utc_seconds, frac_sec)

    def unpack(self, buf):
        """
        Unpack a string buffer into a TimeFormat object
        :param buffer: str
        :return:
        """
        (self.channel_specific_data, s, fs) = struct.unpack("<III", buf)
        if (self.channel_specific_data >> 4) & 0xF != 0: # Make PTP the default
            self.nanoseconds = fs
        else:
            self.nanoseconds = int(fs / (pow(2, 32)/1e9)) # NTP

        self.datetime = datetime.fromtimestamp(s)

        return True

    def __eq__(self, other):
        if not isinstance(other, TimeDataFormat2):
            return False

        _match_att = ("channel_specific_data", "datetime", "nanoseconds")

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        return "TimeFormat2 ChannelSpecificWord={:#0X} Time={} NanoSeconds={}".format(
            self.channel_specific_data, self.datetime.strftime("%H:%M:%S %x %d-%b %Y"), self.nanoseconds
        )

    def __len__(self):
        return 3*4


class ARINC429DataPacket(object):
    """
    Data Packet Format. Contains a list of Arinc Data Words
    
    :type msgcount: int
    :type arincwords: list[ARINC429DataWord]
    
    
    >>> c = Chapter10UDP()
    >>> arinc_p = ARINC429DataPacket()
    >>> arinc_p.unpack(c.chapter10.payload))
    >>> print arinc_p
    ARINCPayload: MessageCount=0
      ARINCData: GapTime=0 FormatError=False ParityError=False BusSpeed=0 Bus=0
    
    """

    def __init__(self):
        self.msgcount = None  #: The number ofARINC-429 words included in the packet.
        self.arincwords = []  #: List of :class:`ARINC429DataWord`

    def pack(self):
        """
        Pack the ARINC-429 data packet object into a binary buffer

        :rtype: str 
        """
        ret_str = struct.pack("<HH", self.msgcount, 0)
        for a in self.arincwords:
            ret_str += a.pack()

        return ret_str

    def unpack(self, buffer):
        """
        Unpack a string buffer into an ARINC-429 data packet object

        :param buffer: A string buffer representing an ARINC-429 data  packet
        :type buffer: str
        :rtype: None
        """
        CH_SPECIFIC_HDR_LEN = 4
        ARINC_WORD_LEN = 8
        (self.msgcount, _res) = struct.unpack_from("<HH", buffer)
        exp_msg = (len(buffer) - CH_SPECIFIC_HDR_LEN) // ARINC_WORD_LEN
        for msg_idx in range(exp_msg):
            offset = (msg_idx * ARINC_WORD_LEN) + CH_SPECIFIC_HDR_LEN
            arinc_data = ARINC429DataWord()
            arinc_msg_word_buffer = buffer[offset:offset+ARINC_WORD_LEN]
            arinc_data.unpack(arinc_msg_word_buffer)
            self.arincwords.append(arinc_data)
            
        if self.msgcount != len(self.arincwords):
            raise Exception("The ARINC Message Count={} does not match number of messages in the packet={}".format(
                self.msgcount, len(self.arincwords)
            ))
        return True

    def __eq__(self, other):
        if not isinstance(other, ARINC429DataPacket):
            return False

        _match_att = ("msgcount", "arincwords")

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        ret_str =  "ARINCPayload: MessageCount={}\n".format( self.msgcount)

        for a in self.arincwords:
            ret_str += "  {}\n".format(repr(a))

        return ret_str

    def __iter__(self):
        self._index = 0
        return self

    def next(self):
        if self._index < len(self.arincwords):
            _dw = self.arincwords[self._index]
            self._index += 1
            return _dw
        else:
            raise StopIteration

    __next__ = next

    def __len__(self):
        return len(self.arincwords)

    def __getitem__(self, key):
        return self.arincwords[key]


class ARINC429DataWord(object):
    """
    The Chapter 10 standard defines specific payload formats for different data. This class handles AROINC-429 packets
    
    :type msgcount: int
    :type gaptime: int
    :type format_error: bool
    :type parity_error: bool
    :type bus_speed: int
    :type bus: int
    :type payload: str
    """

    LO_SPEED = 0  #: Bus speed constant
    HI_SPEED = 1  #: Bus speed constant

    HDR_FORMAT = ">HBB"

    def __init__(self):
        self.gaptime = None  #: The gap time from the beginning of the preceding bus word (regardless of bus) to the
        # beginning of the current bus word in 0.1-us increments
        self.format_error = False  #: Format error has occurred
        self.parity_error = False  #: Parity error has occurred
        self.bus_speed = ARINC429DataWord.LO_SPEED  #: Arinc bus speed
        self.bus = None  #: Bus number index from 0
        self.payload = b""  #: ARINC word as a string payload

    def pack(self):
        """
        Pack the ARINC-429 data packet object into a binary buffer

        :rtype: str|bytes
        """
        _flag = (self.format_error << 7) + (self.parity_error << 6) + (self.bus_speed << 5) + (self.gaptime >> 16)
        _gap = self.gaptime & 0xFFFF
        hdr = struct.pack(ARINC429DataWord.HDR_FORMAT, _gap, _flag, self.bus)

        return hdr + self.payload

    def unpack(self, buffer):
        """
        Unpack a string buffer into an ARINC-429 data packet object

        :param buffer: A string buffer representing an ARINC-429 data  packet
        :type buffer: str
        :rtype: None
        """
        ( _gap, _flag, self.bus) = struct.unpack_from(ARINC429DataWord.HDR_FORMAT, buffer)
        self.payload = buffer[struct.calcsize(ARINC429DataWord.HDR_FORMAT):]
        self.format_error = bool((_flag >> 7) & 0x1)
        self.parity_error = bool((_flag >> 6) & 0x1)
        self.bus_speed = (_flag >> 6) & 0x1
        self.gaptime = (_flag << 16) + _gap

        return True

    def __eq__(self, other):
        if not isinstance(other, ARINC429DataWord):
            return False

        _match_att = ( "gaptime", "format_error", "parity_error", "bus_speed", "bus", "payload")

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        return "ARINCData: GapTime={} FormatError={} ParityError={} BusSpeed={} Bus={}".format(
             self.gaptime, self.format_error, self.parity_error, self.bus_speed, self.bus)


class UARTDataPacket(object):
    """
    Data Packet Format. Contains a list of UART Data Words

    :type msgcount: int
    :type uartwords: list[UARTDataWord]


    >>> c = Chapter10UDP()
    >>> uart_p = UARTDataPacket()
    >>> uart_p.unpack(c.chapter10.payload))
    >>> print uart_p


    """

    def __init__(self):
        self.uartwords = []  #: List of :class:`UARTDataWord`

    def pack(self):
        """
        Pack the UART data packet object into a binary buffer

        :rtype: str
        """
        # Some checks
        if len(self) == 0:
            raise Exception("No UARTDataWords defined")

        # All words should have the same configuraiton of timestamps
        ts_present = None
        for udw in self:
            if ts_present is None:
                ts_present = (udw.ptptimeseconds is not None)
            elif ts_present != (udw.ptptimeseconds is not None):
                raise Exception("All UARTDataWords should either have or have not a PTPTimestamp header")
        ret_buf = struct.pack("<I", int(ts_present) << 31)
        for udw in self:
            ret_buf += udw.pack()

        return ret_buf

    def unpack(self, mybuffer):
        """
        Unpack a string buffer into an UART data packet object

        :param buffer: A string buffer representing an UART data  packet
        :type buffer: str
        :rtype: None
        """

        (ch_spec_word,) = struct.unpack_from("<I", mybuffer)
        ts_present = bool(ch_spec_word >> 31)
        offset = 4
        while abs(offset - len(mybuffer)) > 4:
            udw = UARTDataWord()
            offset += udw.unpack(mybuffer[offset:], ts_present)
            self.uartwords.append(udw)

        return True

    def append(self, udw):
        """
        Add a UART DW to the DP

        :type udw: UARTDataWord
        :rtype: bool
        """
        if not isinstance(udw, UARTDataWord):
            raise Exception("Can only append UARTDataWords")
        return self.uartwords.append(udw)

    def __eq__(self, other):
        if not isinstance(other, UARTDataPacket):
            return False

        _match_att = ["uartwords"]

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        ret_str = "UARTPayload: UARTDataWordCount={}\n".format(len(self))

        for a in self:
            ret_str += "  {}\n".format(repr(a))

        return ret_str

    def __iter__(self):
        self._index = 0
        return self

    def next(self):
        if self._index < len(self.uartwords):
            _dw = self.uartwords[self._index]
            self._index += 1
            return _dw
        else:
            raise StopIteration

    __next__ = next

    def __len__(self):
        return len(self.uartwords)

    def __getitem__(self, key):
        return self.uartwords[key]


class UARTDataWord(object):
    """
    The Chapter 10 standard defines specific payload formats for different data. This class handles UART packets

    :type ptptimeseconds: int
    :type ptptimenanoseconds: int
    :type subchannel: int
    :type parity_error: bool
    :type datalength: int
    :type payload: str
    """

    def __init__(self):
        self.ptptimeseconds = None #: Timestamp of first parameter in the packet. EPOCH time
        self.ptptimenanoseconds = None #: Nanaosecond timestamp
        self.parity_error = False  #: Parity error has occurred
        self.subchannel = None  #: Subchannel
        self.datalength = None  #: Data Length
        self._payload = b""  #: UART payload

    @property
    def payload(self):
        return self._payload

    @payload.setter
    def payload(self, mybuffer):

        self._payload = mybuffer
        self.datalength = len(mybuffer)

    def pack(self):
        """
        Pack the UART data packet object into a binary buffer

        :rtype: str|bytes
        """
        if self.ptptimeseconds is not None and self.ptptimenanoseconds is not None:
            ch_spec_word = struct.pack("<II",self.ptptimenanoseconds, self.ptptimeseconds)
        else:
            ch_spec_word = b""
        data_len = len(self.payload)
        if self.parity_error:
            _subch = self.subchannel + 0x80
        else:
            _subch = self.subchannel
        intra_pkt_header = struct.pack("<HH", data_len, _subch)

        if data_len % 2 == 1:
            padding = struct.pack("B", 0xFF)
        else:
            padding = b""

        return ch_spec_word + intra_pkt_header + self.payload + padding

    def unpack(self, mybuffer, timestamp_present=True):
        """
        Unpack a string buffer into an UART data packet object. Returns the buffer that was consumed

        :param mybuffer: A string buffer representing an UART data  packet
        :type mybuffer: str
        :param timestamp_present: The UART DataPacket Format knows if there is a timestamp or not
        :type timestamp_present: bool
        :rtype: int
        """
        offset = 0
        if timestamp_present:
            #bytes = struct.unpack_from(">8B", mybuffer)
            (self.ptptimenanoseconds, self.ptptimeseconds) = struct.unpack_from("<II", mybuffer, offset)
            offset += 8
        (self.datalength, _pe_sub) = struct.unpack_from("<HH", mybuffer, offset)
        offset += 4

        self.subchannel = _pe_sub & 0x1FFF
        self.parity_error = bool(_pe_sub >> 15)
        self.payload = mybuffer[offset:offset+self.datalength]
        offset += self.datalength

        if self.datalength % 2 == 1:
            return offset + 1
        else:
            return offset

    def __eq__(self, other):
        if not isinstance(other, UARTDataWord):
            return False

        _match_att = ["ptptimeseconds", "ptptimenanoseconds", "parity_error", "subchannel", "datalength", "payload"]

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        return "UARTDataWord: PTPSec={} PTPNSec={} ParityError={} DataLen={} SubChannel={}".format(
             self.ptptimeseconds, self.ptptimenanoseconds, self.parity_error, self.datalength, self.subchannel)


class PCMMinorFrame(object):

    HDR_LEN = 10
    """
    Object that represents the PCM minor frame in a PCMPayload.
    """
    def __init__(self):
        self.intra_packet_sec = 0
        self.intra_packet_nsec = 0
        self.intra_packet_data_header = 0x0
        self.minor_frame_data = b''
        self.syncword = None
        self.sfid = None

    def unpack(self, buffer, extract_sync_sfid=False):
        """
        Convert a string buffer into a PCMDataPacket
        :type buffer: str
        :rtype: bool
        """

        (self.intra_packet_nsec, self.intra_packet_sec, self.intra_packet_data_header) = \
            struct.unpack_from("<IIH", buffer)
        if extract_sync_sfid:
            (self.syncword, self.sfid) = struct.unpack_from(">IH", buffer, 10)
        self.minor_frame_data = buffer[10:]
        return True

    def pack(self):
        """
        Convert a PCMFrame object into a string buffer
        :return:
        """
        buf = struct.pack("<IIH", self.intra_packet_nsec, self.intra_packet_sec, self.intra_packet_data_header)
        if self.syncword is not None:
            buf += struct.pack(">I", self.syncword)
        if self.sfid is not None:
            buf += struct.pack(">H", self.sfid)
        buf += self.minor_frame_data

        return buf

    def __repr__(self):
        time_fmt = "%H:%M:%S %d-%b %Y"
        date_str = datetime.fromtimestamp(self.intra_packet_sec).strftime(time_fmt)
        return "Minor Frame. Sec={} ({}) NanoSec={} DataHdr={:#0X} ".format(
            self.intra_packet_sec, date_str, self.intra_packet_nsec, self.intra_packet_data_header
        )

    def __eq__(self, other):
        if not isinstance(other, PCMMinorFrame):
            return False
        for attr in ["intra_packet_sec", "intra_packet_nsec", "intra_packet_data_header", "minor_frame_data",
                     "syncword", "sfid"]:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

PCM_DATA_FRAME_FILL = 0x0


class PCMDataPacket(object):
    """
    This object represents the Payload to a Chapter 10 PCM packet
    The user needs to tell the object how many minor frames in Payload before unpacking a buffer.
    """
    def __init__(self):
        self.channel_specific_word = None
        self.minor_frame_size_bytes = None
        self.minor_frames = []

    def unpack(self, buffer):
        """
        Convert a string buffer into a PCMDataPacket
        :type buffer: bytes
        :rtype: bool
        """

        (self.channel_specific_word,) = struct.unpack_from("<I", buffer, 0)
        offset = 4
        while offset < len(buffer):
            minor_frame = PCMMinorFrame()
            if (self.minor_frame_size_bytes + PCMMinorFrame.HDR_LEN) % 2 != 0:
                padding = 1
            else:
                padding = 0
            minor_frame.unpack(buffer[offset:offset+self.minor_frame_size_bytes+10])
            offset += (self.minor_frame_size_bytes+10+padding)
            self.minor_frames.append(minor_frame)

        return True

    def pack(self):
        buf = struct.pack("<I", self.channel_specific_word)
        for mf in self.minor_frames:
            buf += mf.pack()
            if len(mf.pack()) % 2 == 1:
                buf += struct.pack(">B", PCM_DATA_FRAME_FILL)

        return buf

    def __repr__(self):
        _rstr = "PCM Data Packet Format 1. Channel Specific Word ={:#0X}\n".format(self.channel_specific_word)
        for m in self.minor_frames:
            _rstr += "{}\n".format(repr(m))

        return _rstr

    def __iter__(self):
        self._index = 0
        return self

    def next(self):
        if self._index < len(self.minor_frames):
            _frame = self.minor_frames[self._index]
            self._index += 1
            return _frame
        else:
            raise StopIteration

    __next__ = next

    def __eq__(self, other):
        """

        :type other: PCMDataPacket
        :return:
        """
        if not isinstance(other, PCMDataPacket):
            return False

        if self.channel_specific_word != other.channel_specific_word:
            return False

        if len(self.minor_frames) != len(other.minor_frames):
            return False

        for idx in range(len(self.minor_frames)):
            if self.minor_frames[idx] != other.minor_frames[idx]:
                return False

        return True


class MILSTD1553DataPacket(object):
    """
    Data Packet Format. Contains a list of MIML-STD-1553 Data Words

    :type msgcount: int
    :type uartwords: list[MILSTD1553Message]


    >>> c = Chapter10UDP()
    >>> m = MILSTD1553DataPacket()
    >>> m.unpack(c.chapter10.payload))
    >>> print m


    """

    def __init__(self):
        self.messages = []  #: List of :class:`MILSTD1553Message`
        self.msgcount = 0
        self.ttb = 0

    def pack(self):
        """
        Pack the MIL-STD-1553 data packet object into a binary buffer

        :rtype: str
        """
        # Some checks
        if len(self) == 0:
            raise Exception("No MILSTD1553Messages defined")

        msg_buf = bytes()
        for dw in self:
            msg_buf += dw.pack()

        csw = struct.pack("<I", (self.ttb << 30)+len(self))

        return csw + msg_buf

    def unpack(self, mybuffer):
        """
        Unpack a string buffer into an UART data packet object

        :param buffer: A string buffer representing an UART data  packet
        :type buffer: str
        :rtype: None
        """

        (ch_spec_word,) = struct.unpack_from("<I", mybuffer)
        self.msgcount = ch_spec_word & 0xFFFFFF
        self.ttb = (ch_spec_word >> 30) & 0x3

        offset = 4
        while offset+14 < len(mybuffer):  # Should have at least the timestamp
            m = MILSTD1553Message()
            offset += m.unpack(mybuffer[offset:])
            self.messages.append(m)

        return True

    def append(self, message):
        """
        Add a message to the DP

        :type message: MILSTD1553Message
        :rtype: bool
        """
        if not isinstance(message, MILSTD1553Message):
            raise Exception("Can only append MILSTD1553Message")
        self.msgcount += 1
        return self.messages.append(message)

    def __eq__(self, other):
        if not isinstance(other, MILSTD1553DataPacket):
            return False

        _match_att = ["messages", "msgcount"]

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        ret_str = "MILSTD1553DataPacket: MessageCount={}\n".format(self.msgcount, self.messages)

        for a in self:
            ret_str += "  {}\n".format(repr(a))

        return ret_str

    def __iter__(self):
        self._index = 0
        return self

    def next(self):
        if self._index < len(self.messages):
            _dw = self.messages[self._index]
            self._index += 1
            return _dw
        else:
            raise StopIteration

    __next__ = next

    def __len__(self):
        return len(self.messages)

    def __getitem__(self, key):
        return self.messages[key]


class MILSTD1553Message(object):
    """
    The Chapter 10 standard defines specific payload formats for different data. This class handles UART packets

    :type ptptimeseconds: int
    :type ptptimenanoseconds: int
    :type subchannel: int
    :type parity_error: bool
    :type datalength: int
    :type payload: str
    """

    def __init__(self):
        self.ptptimeseconds = None #: Timestamp of first parameter in the packet. EPOCH time
        self.ptptimenanoseconds = None #: Nanaosecond timestamp
        self.blockstatus = 0
        self.gaptimes = 0
        self.length = 0
        self.message = None

    def pack(self):
        """
        Pack the MIL-STD-1553 message object into a binary buffer

        :rtype: str|bytes
        """
        if self.ptptimeseconds is not None and self.ptptimenanoseconds is not None:
            ch_spec_word = struct.pack("<II",self.ptptimenanoseconds, self.ptptimeseconds)
        else:
            raise Exception("No timestamp defined")
        self.length = len(self.message)
        intra_packet_data_header = struct.pack("<HHH", self.blockstatus, self.gaptimes, self.length)

        return ch_spec_word + intra_packet_data_header + self.message

    def unpack(self, mybuffer):
        """
        Unpack a string buffer into an MIL-STD-1553 data packet object. Returns the buffer that was consumed

        :param mybuffer: A string buffer representing an UART data  packet
        :type mybuffer: str
        :rtype: int
        """
        offset = 0
        #bytes = struct.unpack_from(">8B", mybuffer)
        (self.ptptimenanoseconds, self.ptptimeseconds) = struct.unpack_from("<II", mybuffer, offset)
        offset += 8
        (self.blockstatus, self.gaptimes, self.length) = struct.unpack_from("<HHH", mybuffer, offset)
        offset += 6

        self.message = mybuffer[offset:offset+self.length]
        offset += self.length

        return offset

    def __eq__(self, other):
        if not isinstance(other, MILSTD1553Message):
            return False

        _match_att = ["ptptimeseconds", "ptptimenanoseconds", "blockstatus", "gaptimes", "length", "message"]

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        return "MILSTD1553Message: PTPSec={} PTPNSec={} BlockStatus={} GapTimes={} Length={}".format(
             self.ptptimeseconds, self.ptptimenanoseconds, self.blockstatus, self.gaptimes, self.length)



#-------------------------------------------------------------------------------
# Name:        pcap_monitor.py
# Purpose:
#
# Author:      Diarmuid
#
# Created:     16/12/2013
# Copyright:   (c) ACRA 2013
# Licence:     <your licence>
# About:       This is a simple test script that loops through a pcap file pulling out the
#               packets.
#-------------------------------------------------------------------------------

import socket,os,struct,sys
import argparse
import datetime, time


import AcraNetwork.iNetX as inetx
import AcraNetwork.Pcap as pcap
import AcraNetwork.SimpleEthernet as SimpleEthernet
import AcraNetwork.ParserAligned as ParserAligned




def main():

    try:
        pcapfile = pcap.Pcap("SSR_ABM_102_capture_example1.pcap")
    except IOError:
        print("ERROR: Could not find input file SSR_ABM_102_capture_example1.pcap")
        exit()

    # Keep a running count of the packets
    packet_count = 1
    # Keep a count of previous sequence number to detect a dropped packets
    PreviousSeqNum = dict()

    while True: # while we are not at the end of the file
        try:
            # So we loop through the file one packet at a time. This will eventually return an
            # exception at the end of file so handle that when it occurs
            pcaprecord = pcapfile.readAPacket()
            eth = SimpleEthernet.Ethernet()
            eth.unpack(pcaprecord.packet)
            ip = SimpleEthernet.IP()
            ip.unpack(eth.payload)
            udp_packet = SimpleEthernet.UDP()
            udp_packet.unpack(ip.payload)
            (ctrl_word,) = struct.unpack('>I',udp_packet.payload[:4])
            if ctrl_word == 0x11000000: # This is a rough guess assuming the control word is 0x11000000
                inetx_packet = inetx.iNetX()
                inetx_packet.unpack(udp_packet.payload)

                #----------------------------
                # Check for dropped packet
                #----------------------------
                if inetx_packet.streamid not in PreviousSeqNum:
                    PreviousSeqNum[inetx_packet.streamid] = inetx_packet.sequence
                else:
                    if PreviousSeqNum[inetx_packet.streamid]+1 != inetx_packet.sequence:
                        print("ERROR: Dropped {} packets on streamid={:#x} at packet count={}".format(inetx_packet.sequence - PreviousSeqNum[inetx_packet.streamid] + 1,inetx_packet.streamid,packet_count))
                    PreviousSeqNum[inetx_packet.streamid] = inetx_packet.sequence

                print("----- StreamID={:#10x} SourceIP= {:10s} -----".format(inetx_packet.streamid,ip_packet.srcip))
                #--------------------------------------------------------------------------------
                # Packets on stream id 0x11121314 is a parser aligned block so lets look at this
                #--------------------------------------------------------------------------------
                if inetx_packet.streamid == 0x11121314:
                    parser_aligned_packet = ParserAligned.ParserAlignedPacket()
                    # unpack the _payload as the parser data
                    parser_aligned_packet.unpack(inetx_packet.payload)

                    # Loop through all the blocks in the packet and spit them out
                    for pblock in parser_aligned_packet.parserblocks:
                        (payload_data,) =struct.unpack('>I',pblock.payload)
                        print("Sequence Number = {:8} Quadbyes={:5} Msgcnt={:5} BusId={:4} Elapsed={:20} ".format(inetx_packet.sequence, pblock.quadbytes,pblock.messagecount,pblock.busid,pblock.elapsedtime,payload_data))


            packet_count += 1

        except NotImplementedError:
            # We received a packet that we don't care about. So skip silently
            packet_count += 1
            pass

        except IOError:
            # We are at the end of the file so lets jump to the next file
            print ( "End of file reached")
            exit()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
===== 
Multicast reception example
===== 


"""
__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__version__ = "0.1.0"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"

import sys
sys.path.append("..")

import socket
import AcraNetwork.iNetX as inetx
import AcraNetwork.McastSocket as McastSocket

udp_port = 5567

#------------------------------------------------------------
# Setup a socket to recieve all traffic
#------------------------------------------------------------
try:
    recv_socket = McastSocket.McastSocket(local_port=udp_port, reuse=1)
    recv_socket.mcast_add('235.0.0.1', '0.0.0.0')
    recv_socket.settimeout(10)
except:
    print("Can't bind to socket {}".format(udp_port))
    exit()

packet_count = 1
while True:
    try:
        data, addr = recv_socket.recvfrom(2048) # buffer size is 1500 bytes
    except socket.timeout:
        print("timeout on socket")
        exit()

    (udpsrcport,srcipaddr) = addr
    avionics_packet = inetx.iNetX()
    data_len = len(data)
    try:
        avionics_packet.unpack(data)
    except ValueError:
        # This isn't an inetx packet
        packet_count += 1
        continue
    else:
        print("Packet withStread ID = {:0X}".format(avionics_packet.streamid))


#-------------------------------------------------------------------------------
# Name:        
# Purpose:     
#
# Author:      
#
# Created:     
#
__author__ = '$USER'

import sys
sys.path.append("..")

import unittest
import datetime
import time
import AcraNetwork.IENA as iena
import AcraNetwork.SimpleEthernet as SimpleEthernet
import AcraNetwork.Pcap as pcap
import struct
import os
from copy import copy


THIS_DIR = os.path.dirname(os.path.abspath(__file__))


def get_udp_packet(pcapfilename):
    """
    Convienence method to read the first UDP packet in a pcap file

    :param pcapfilename: pcapfilename
    :type pcapfilename: str
    :rtype: SimpleEthernet.UDP
    """
    p = pcap.Pcap(os.path.join(THIS_DIR, pcapfilename))
    p._read_global_header()
    mypcaprecord = p[0]
    e = SimpleEthernet.Ethernet()
    e.unpack(mypcaprecord.packet)
    ip = SimpleEthernet.IP()
    ip.unpack(e.payload)
    udp_pkt = SimpleEthernet.UDP()
    udp_pkt.unpack(ip.payload)
    p.close()

    return udp_pkt

ienq_q="""IENAQ: KEY=0XDC SEQ=2 TIMEUS=0 NUM_QPARAM=5
 Q-Param #0:ParamID=0XA Dataset Length=15
 Q-Param #1:ParamID=0XB Dataset Length=16
 Q-Param #2:ParamID=0XC Dataset Length=17
 Q-Param #3:ParamID=0XD Dataset Length=18
 Q-Param #4:ParamID=0XE Dataset Length=19
"""
class IENATest(unittest.TestCase):

    def setUp(self):
        # Get aUDP packet from a pack which can be used
        p = pcap.Pcap(os.path.join(THIS_DIR, "iena_test.pcap"))
        p._read_global_header()

        # Loop through the pcap file reading one packet at a time
        try:
            mypcaprecord = p[0]
        except IOError:
            # End of file reached
            return
        e = SimpleEthernet.Ethernet()
        e.unpack(mypcaprecord.packet)
        ip =  SimpleEthernet.IP()
        ip.unpack(e.payload)
        self.upd = SimpleEthernet.UDP()
        self.upd.unpack(ip.payload)
        p.close()


    def test_defaultIENA(self):
        '''Check the defaults for the IENA packet'''
        i = iena.IENA()
        self.assertEqual(i.key,None)
        self.assertEqual(i.size,None)
        self.assertEqual(i.keystatus,None)
        self.assertEqual(i.status,None)
        self.assertEqual(i.sequence,None)
        self.assertEqual(i.endfield,0xdead)
        self.assertEqual(i.payload,None)

    def test_basicIENA(self):
        '''Build a very basic packet and check the packed result'''
        i = iena.IENA()
        i.key = 1
        i.keystatus = 2
        i.status = 3
        i.sequence = 10
        i.setPacketTime(time.mktime(datetime.datetime(datetime.datetime.today().year, 1, 2, 0, 0, 0,0).timetuple()),0)
        i.payload = struct.pack('H',0x5)
        # size = 9. Time = midnight Jan 02 = 86400s 0us = 0x14dd760000
        expected_payload = struct.pack(iena.IENA.IENA_HEADER_FORMAT,1,9,0x14,0x1dd76000,2,3,10) + struct.pack('H',0x5) + struct.pack('>H',0xdead)
        self.assertEqual(i.pack(),expected_payload)

    def test_basicIENA_streamid(self):
        '''Build a very basic packet and check the packed result'''
        i = iena.IENA()
        i.streamid = 1
        i.keystatus = 2
        i.status = 3
        i.sequence = 10
        self.assertEqual(i.key, i.streamid)
        self.assertEqual(i.status, i.n2)
        i.setPacketTime(time.mktime(datetime.datetime(datetime.datetime.today().year, 1, 2, 0, 0, 0,0).timetuple()),0)
        i.payload = struct.pack('H',0x5)
        # size = 9. Time = midnight Jan 02 = 86400s 0us = 0x14dd760000
        expected_payload = struct.pack(iena.IENA.IENA_HEADER_FORMAT,1,9,0x14,0x1dd76000,2,3,10) + struct.pack('H',0x5) + struct.pack('>H',0xdead)
        self.assertEqual(i.pack(),expected_payload)
        self.assertEqual(repr(i), "IENAP: KEY=0X1 SEQ=10 TIMEUS=86400000000")

    def test_unpackIEANFromPcap(self):
        '''Read all the IENA packets in a pcap file and check each field'''

        # Now I have a _payload that will be an iena packet
        i = iena.IENA()
        i.unpack(self.upd.payload)
        self.assertEqual(i.key,0x1a)
        self.assertEqual(i.size,24)
        self.assertEqual(i.status,1)
        self.assertEqual(i.keystatus,1)
        self.assertEqual(i.sequence, 195)
        self.assertEqual(i.timeusec,0x1d102f800)
        self.assertEqual(i.endfield,0xdead)
        self.assertEqual(len(i), 48)
        i2 = copy(i)
        self.assertTrue(i2 == i)

    def test_unpack_IENAM(self):
        '''Read all the IENA packets in a pcap file and check each field'''
        # Now I have a _payload that will be an iena packet
        i = iena.IENAM()
        i.unpack(self.upd.payload)
        self.assertEqual(i.key,0x1a)
        for mparam in i:
            self.assertEqual(mparam.paramid, 0xDC)
            self.assertEqual(repr(mparam), "ParamID=0XDC Delay=16 Dataset Length=26")
        #print(i)
        self.assertEqual(repr(i), "IENAM: KEY=0X1A SEQ=195 TIMEUS=7801600000 NUM_MPARAM=1\n M-Param #0:ParamID=0XDC Delay=16 Dataset Length=26\n")

    def test_unpack_IENAN(self):
        # Decode as IENA-N
        i = iena.IENAN()
        i.unpack(self.upd.payload)
        self.assertEqual(i.key, 0x1a)
        param_1 = i.parameters[0] # type: iena.NParameter
        self.assertEqual(param_1.paramid, 0xDC)
        self.assertListEqual(param_1.dwords, [0x10])
        self.assertEqual(repr(i), "IENAN: KEY=0X1A SEQ=195 TIMEUS=7801600000 NUM_DPARAM=8")
        self.assertEqual(len(i), 8)
        for idx, n in enumerate(i):
            if idx == 0:
                self.assertEqual(n.paramid, 220)

    def test_unpack_IENAD(self):
        # Decode as IENA-N
        udp_pkt = get_udp_packet("ienad.pcap")
        i = iena.IENAD()
        i.unpack(udp_pkt.payload)
        self.assertEqual(i.key, 0x2cfa)
        param_1 = i.parameters[0] # type: iena.DParameter
        self.assertEqual(param_1.paramid, 0xFFFF)
        self.assertEqual(param_1.delay, 0x0)
        self.assertListEqual(param_1.dwords, [0xfed1, 0x7cfe])
        self.assertEqual(repr(i), "IENAD: KEY=0X2CFA SEQ=0 TIMEUS=1837 NUM_DPARAM=11")
        self.assertEqual(len(i), 11)
        for id,p in enumerate(i):
            if id == 0:
                self.assertEqual(p.paramid, 0xFFFF)
        self.assertEqual(i[0].paramid, 0xffff)


    def test_corrupt_IENAD(self):
        udp_pkt = get_udp_packet("corrupt_ienad.pcap")
        iena_pkt = iena.IENAD()
        self.assertRaises(Exception, lambda : iena_pkt.unpack(udp_pkt.payload) )

    def test_create_ienam(self):
        i = iena.IENAM()
        i.key = 0xDC
        i.endfield = 0xDEAD
        i.keystatus = 0
        i.sequence = 2
        i.n2 = 0
        for idx in range(5):
            mparam = iena.MParameter(paramid=idx, delay=idx*2, dataset=os.urandom(idx+5))
            i.parameters.append(mparam)

        buf = i.pack()
        b = iena.IENAM()
        b.unpack(buf)
        self.assertEqual(b.sequence, 2)
        self.assertEqual(len(b), 5)
        for idx, p in enumerate(b):
            self.assertEqual(p.paramid, idx)
            self.assertEqual(p.delay, idx*2)
            self.assertEqual(len(p.dataset), idx+5)

    def test_create_ienaq(self):
        i = iena.IENAQ()
        i.key = 0xDC
        i.endfield = 0xDEAD
        i.keystatus = 0
        i.sequence = 2
        i.n2 = 0

        for idx in range(10,15):
            qparam = iena.QParameter(paramid=idx, dataset=struct.pack(">{}B".format(idx+5), *range(idx+5)))
            if idx == 10:
                #print repr(qparam)
                self.assertEqual(repr(qparam), "ParamID=0XA Dataset Length=15")

            i.parameters.append(qparam)
        self.assertEqual(repr(i), ienq_q)
        buf = i.pack()
        b = iena.IENAQ()
        b.unpack(buf)
        self.assertEqual(b.sequence, 2)
        self.assertEqual(len(b), 5)
        for idx, p in enumerate(b):
            self.assertEqual(p.paramid, idx+10)
            self.assertEqual(len(p.dataset), idx+5+10)
        #print(i)

    def test_read_multiple_ienam(self):
        # Get aUDP packet from a pack which can be used
        p = pcap.Pcap(os.path.join(THIS_DIR, "ienam_multiple.pcap"))

        # Loop through the pcap file reading one packet at a time
        mypcaprecord = p[0]
        e = SimpleEthernet.Ethernet()
        e.unpack(mypcaprecord.packet)
        ip = SimpleEthernet.IP()
        ip.unpack(e.payload)
        udp = SimpleEthernet.UDP()
        udp.unpack(ip.payload)

        i = iena.IENAM()
        i.unpack(udp.payload)
        self.assertEqual(len(i), 2)
        p.close()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
===== 
 Sending UDP packets
===== 

Send UDP packets at a specific rate
"""
__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__version__ = "0.0.5"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import sys
sys.path.append("..")
import time
import struct
import AcraNetwork.iNetX as inetx
import argparse
import socket
import random
import threading
import queue


parser = argparse.ArgumentParser(description='Send iNetX packets at a specified rate')
parser.add_argument('--rate',required=False, type=int, default=1, help="Packet rate in Mbps")
parser.add_argument('--ipaddress',required=False, type=str, default="192.168.0.26", help="Destination IP")
parser.add_argument('--sidcount',required=False, type=int, default=1, help="number of stream ids to send")
args = parser.parse_args()

# simple application that tests building and sending of iena and inetx packets

UDP_IP = args.ipaddress
UDP_PORT = 4444


def send_to_network(dataq):
    """

    :type dataq: queue.Queue
    :return:
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
    sock.setblocking(True)
    while True:
        payload = dataq.get(block=True, timeout=5)
        sock.sendto(payload, (UDP_IP, UDP_PORT))

sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
#sock.mcast_add(UDP_IP)

# Create an inetx packet
myinetx = inetx.iNetX()

payload = struct.pack(">B",0xa5)
sockets = {}
payload_pkts = {}
bsid = random.randint(0x0, 0xFF) << 8
for idx in range(args.sidcount):
    myinetx.inetxcontrol = inetx.iNetX.DEF_CONTROL_WORD
    myinetx.pif = 0
    myinetx.streamid = bsid + idx
    myinetx.sequence = 0
    myinetx.payload = payload * int(1430 * random.betavariate(3, 1))
    print("Sid={:#0X} Len={}".format(myinetx.streamid, len(myinetx.payload)))
    myinetx.setPacketTime(int(time.time()))
    packet_payload = myinetx.pack()
    payload_pkts[myinetx.streamid] = {'payload': packet_payload, 'length': len(packet_payload) + 8 + 20 + 14}
    pkt_len = len(packet_payload) + 8 + 20 + 14

ave_pkt_len = 1000

pps_rate = int(((args.rate * 1024 * 1024) / (ave_pkt_len * 8)) / 100) * 100
if pps_rate < 1:
    pps_rate = 1
burst_length = args.rate // 4
packet_count = 1
data_vol = 0
dly_ms = int((float(burst_length) * ave_pkt_len) / pps_rate)
delta_change_ms = 5

st = time.time()

print("UDP target IP:", UDP_IP)
print("UDP target port:", UDP_PORT)
print("Rate = {} Hz".format(pps_rate))
print("Rate = {} Mbps".format(args.rate))
print("DLY = {} ms".format(dly_ms))

SEQ_ROLL_OVER = pow(2, 64)
pkt_count = {}
for sid in payload_pkts.keys():
    pkt_count[sid] = 0

while True:
    random_sid = bsid + (packet_count % args.sidcount)
    #random_sid = random.randint(0, args.sidcount-1) + bsid
    random_payload = payload_pkts[random_sid]["payload"]
    # Faster way to build an inetx packet instead of packing the whole header
    mypayload = random_payload[:8] + struct.pack(">I", pkt_count[random_sid]) + random_payload[12:]
    pkt_count[random_sid] += 1
    #payloadq.put(mypayload)
    sock.sendto(mypayload, (UDP_IP, UDP_PORT))
    data_vol += (payload_pkts[random_sid]["length"] * 8)
    if packet_count % burst_length == burst_length-1:
        # Report some information
        if dly_ms > 0:
            time.sleep(dly_ms/1e3)
        ct = time.time()
        rate = data_vol / (ct - st) / 1000 / 1000
        pps = int(packet_count // (ct - st))
        #granularity = pps_rate // 10
        error_val = abs(rate - args.rate)
        if packet_count % (burst_length*10) == burst_length-1:
            print("Rate = {:5.1f} Mbps {:6d} pps Dly={:4d}ms Error={:4.1f}".format(rate, pps, dly_ms, error_val))
        # Tweak the delay so we converge to the required pps
        if rate > args.rate:
            dly_ms += int(delta_change_ms * error_val)
        elif rate < args.rate and dly_ms > 0:
            dly_ms -= int(delta_change_ms * error_val)

    packet_count += 1





#!/usr/bin/env python

from distutils.core import setup
import shutil
from AcraNetwork.__version__ import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='AcraNetwork',
      version=__version__,
      description='Classes and utilities to support Flight Test Instrumentation Ethernet networks',
      author='Diarmuid Collins',
      author_email='dcollins@curtisswright.com',
      url='https://github.com/diarmuidcwc/AcraNetwork',
      packages=['AcraNetwork'],
      scripts=['examples/tx_inetx_udp.py', 'examples/validate_pcap.py'],
      long_description="A collection of classes that can be used to decom network or PCM based FTI traffic",
      classifiers =['Programming Language :: Python',
                    'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
                    'Operating System :: OS Independent',
                    'Development Status :: 3 - Alpha',
                    'Intended Audience :: Developers',
                    'Topic :: Software Development :: Libraries :: Python Modules',
                    ],
     )



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
===== 
Parse MPEG TS packets in pcap....
===== 

"""
__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__version__ = "0.0.1"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import sys
sys.path.append("..")
import AcraNetwork.Pcap as pcap
import AcraNetwork.iNetX as inetx
import AcraNetwork.MPEGTS as mpegts
import AcraNetwork.SimpleEthernet as eth
from AcraNetwork.MPEGTS import H264
import datetime

# This script shows how to parse either a pcap file or a TS file into the constituent
# NALS and finds the unique STANAG SEI User data with the timestamp


def pcap_to_ts(pcapfile,ts_file,udp_port=8010):
    '''
    Convert a pcap file to a TS file by extracting all data from a specified port
    :param mpegfile: str
    :param ts_file: str
    :param udp_port: int
    :return:
    '''

    mpegpcap = pcap.Pcap(pcapfile, mode='r')

    ts = open(ts_file, mode='wb')
    mpeghex = ""
    rec_count = 0

    for rec in mpegpcap:
        try:
            e = eth.Ethernet()
            e.unpack(rec.packet)
            i = eth.IP()
            i.unpack(e.payload)
            u = eth.UDP()
            u.unpack(i.payload)
            if u.dstport == udp_port:
                rec_count += 1
                inet = inetx.iNetX()
                inet.unpack(u.payload)
                mpegtspackets = mpegts.MPEGTS()
                mpegtspackets.unpack(inet.payload)
                for packet in mpegtspackets.blocks:
                    mpeghex += packet.payload
                ts.write(inet.payload)
        except:
            continue


def parse_ts_file(tsfile):

    ts_file = open(tsfile,mode='rb')
    h264_data = H264()
    h264_data.unpack(ts_file.read())

    nal_counts = {}
    timestamp_count = 0
    for nal in h264_data.nals:

        if not nal.type in nal_counts:
            nal_counts[nal.type] = 1
        else:
            nal_counts[nal.type] += 1

        if nal.type == mpegts.NAL_TYPES["SEI"]:
            if nal.sei.unregdata:
                print("Timestamp={} byte offset={} count ={}".format(datetime.datetime.strftime(nal.sei.time,"%d %b %Y %H:%M:%S.%f"),nal.offset,timestamp_count))
                timestamp_count += 1

        else:
            pass

    print("\n----- SUMMARY -----")
    for type in nal_counts:
        print("{} {} NALs in input".format(nal_counts[type],mpegts.NAL_TYPES_INV[type]))
    print("{} STANAG Timestamps".format(timestamp_count))


def main():
    # Read in a TS file and print out some useful information
    parse_ts_file("../test/stanag_sample.ts")


if __name__ == "__main__":
    main()


__author__ = 'diarmuid'
import sys
sys.path.append("..")
import unittest
import AcraNetwork.Pcap as pcap
import AcraNetwork.SimpleEthernet as SimpleEthernet
import AcraNetwork.NPD as NPD
import struct
import os

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


def getEthernetPacket(data=""):
    e = SimpleEthernet.Ethernet()
    e.srcmac = 0x001122334455
    e.dstmac = 0x998877665544
    e.type = SimpleEthernet.Ethernet.TYPE_IP
    i = SimpleEthernet.IP()
    i.dstip = "235.0.0.1"
    i.srcip = "192.168.1.1"
    i.protocol = SimpleEthernet.IP.PROTOCOLS["UDP"]
    u = SimpleEthernet.UDP()
    u.dstport = 6667
    u.srcport = 6667
    u.payload = data
    i.payload = u.pack()
    e.payload = i.pack()
    return e.pack()


test_npd_exp="""NPD: DataType=0X2 Seq=10 DataSrcID=0X1 MCastAddr=235.0.0.1
\tNPD Segment. TimeDelta=3 Segment Len=18 ErrorCode=0 Flags=0X1
\tNPD Segment. TimeDelta=3 Segment Len=22 ErrorCode=0 Flags=0X1
\tNPD Segment. TimeDelta=3 Segment Len=26 ErrorCode=0 Flags=0X1
\tNPD Segment. TimeDelta=3 Segment Len=30 ErrorCode=0 Flags=0X1"""


class testNPD(unittest.TestCase):

    def  setUp(self):
        self.n = NPD.NPD()
        self.n.datasrcid = 1
        self.n.datatype = 2
        self.n.cfgcnt = 3
        self.n.flags = 4
        self.n.timestamp = 5
        self.n.sequence = 10
        self.n.mcastaddr = "235.0.0.1"

        self.ns = NPD.NPDSegment()
        self.ns.timedelta = 3
        self.ns.errorcode = 0
        self.ns.flags = 1

        pcapr = pcap.Pcap(os.path.join(THIS_DIR, "npd_ref.pcap"), mode="r")
        rec = pcapr[0]
        pcapr.close()
        self.readnpd_payload = rec.payload[(14+20+8):]

    def test_npds_print(self):
        self.ns.payload = struct.pack(">II", 2, 3)
        self.assertEqual(repr(self.ns), "NPD Segment. TimeDelta=3 Segment Len=16 ErrorCode=0 Flags=0X1")

    def test_npds_eq(self):
        self.ns.payload = struct.pack(">II", 2, 3)
        n2 = NPD.NPDSegment()
        n2.timedelta = 3
        n2.errorcode = 0
        n2.flags = 1
        n2.payload = struct.pack(">II", 2, 3)
        self.assertTrue(n2 == self.ns)
        n2.flags = 3
        self.assertTrue(n2 != self.ns)

    def test_npd_basic(self):
        # No _payload
        buf=self.n.pack()
        self.assertEqual(len(buf), 20)

    def test_npd_sec(self):
        self.ns.payload =  struct.pack(">III",1,2,3)
        buf = self.ns.pack()
        self.assertEqual(len(buf), 20)

    def test_npd_in_npd(self):
        # Add a segment to the _payload
        self.ns.payload = struct.pack(">III", 1, 2, 3)
        self.n.segments.append(self.ns)
        buf = self.n.pack()
        self.assertEqual(len(buf), 40)

    def test_nd_to_pcap(self):
        for segment in range(3,7):
            ns = NPD.NPDSegment()
            ns.timedelta = 3
            ns.errorcode = 0
            ns.flags = 1
            ns.payload = struct.pack(">{}IH".format(segment-1), *list(range(segment)))
            self.n.segments.append(ns)
        rec = getEthernetPacket(self.n.pack())
        self.pcapw = pcap.Pcap("test_npd.pcap", mode="w")
        self.pcapw.write_global_header()
        self.rec = pcap.PcapRecord()
        self.rec.payload = rec
        self.pcapw.write(self.rec)
        self.pcapw.close()

    def test_odd_payload(self):
        # Add a segment to the _payload
        self.ns.payload = struct.pack(">IIH", 1, 2, 3)
        self.n.segments.append(self.ns)
        buf = self.n.pack()
        self.assertEqual(len(buf), 40)

    def test_unpack(self):
        n = NPD.NPD()
        n.unpack(self.readnpd_payload)
        self.assertEqual(len(n),4)
        self.assertEqual(n.segments[3].segmentlen, 30)
        #print repr(n)
        self.assertEqual(repr(n), test_npd_exp)
        for i,s in enumerate(n):
            if i == 1:
                self.assertEqual(s.timedelta, 3)

    @unittest.skip("no pcap")
    def test_unpackwrap(self):
        pcapr = pcap.Pcap(os.path.join(THIS_DIR, "npd_pcm.pcap"), mode="r")
        rec = pcapr[0]
        readnpd_payload = rec.payload[(14+20+8):]
        n = NPD.NPD()
        n.unpack(readnpd_payload)
        self.assertEqual(len(n.segments),2)
        self.assertEqual(n.segments[0].segmentlen, 34)
        #print(repr(n))
        self.assertEqual(n.segments[0].sfid, 0x0)
        pcapr.close()

    def test_rs232(self):
        self.n.datatype = 0x50
        for segment in range(3, 7):
            ns = NPD.RS232Segment()
            ns.timedelta = 3
            ns.errorcode = 0
            ns.flags = 1
            ns.block_status = NPD.RS232Segment.BSL_PARN_EN + NPD.RS232Segment.BSL_422
            ns.data = os.urandom(50+segment)
            self.n.segments.append(ns)
        for segment in range(1, 5):
            ns = NPD.RS232Segment()
            ns.timedelta = 3
            ns.errorcode = 0
            ns.flags = 1
            ns.block_status = NPD.RS232Segment.BSL_PARN_EN + NPD.RS232Segment.BSL_422
            ns.sync_bytes = [segment] * segment
            ns.data = os.urandom(50+segment)
            self.n.segments.append(ns)

        rec = getEthernetPacket(self.n.pack())
        self.pcapw = pcap.Pcap("test_npd_rs232.pcap", mode="w")
        self.pcapw.write_global_header()
        self.rec = pcap.PcapRecord()
        self.rec.payload = rec
        self.pcapw.write(self.rec)
        self.pcapw.close()

        n2= NPD.NPD()
        n2.unpack(self.n.pack())
        self.assertEqual(self.n, n2)
        self.assertEqual(repr(n2[0]), "RS232 NPD Segment. TimeDelta=3 Segment Len=63 ErrorCode=0X0 Flags=0X1 Block_Status=0X840 DataLen=53")


if __name__ == '__main__':
    unittest.main()



__author__ = 'diarmuid'
import sys
sys.path.append("..")

import unittest
import AcraNetwork.iNetX as inetx
import AcraNetwork.SimpleEthernet as SimpleEthernet
import AcraNetwork.Pcap as pcap
import struct
import os
import copy

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

class iNetXTest(unittest.TestCase):

    def test_defaultiNet(self):
        i = inetx.iNetX()
        self.assertEqual(i.packetlen,None)
        self.assertEqual(i.inetxcontrol,inetx.iNetX.DEF_CONTROL_WORD)
        self.assertEqual(i.ptptimeseconds,None)
        self.assertEqual(i.ptptimenanoseconds,None)
        self.assertEqual(i.pif,None)
        self.assertEqual(i.payload,None)
        self.assertEqual(i.sequence,None)
        self.assertEqual(i.streamid,None)

    def test_basiciNet(self):
        i = inetx.iNetX()
        i.sequence = 1
        i.pif = 0
        i.streamid = 0xdc
        i.setPacketTime(1,1)
        i.payload = struct.pack('H',0x5)
        expected_payload = struct.pack(inetx.iNetX.INETX_HEADER_FORMAT,inetx.iNetX.DEF_CONTROL_WORD,0xdc,1,30,1,1,0) + struct.pack('H',0x5)
        self.assertEqual(i.pack(),expected_payload)

    def test_unpackiNet(self):
        expected_payload = struct.pack(inetx.iNetX.INETX_HEADER_FORMAT,inetx.iNetX.DEF_CONTROL_WORD, 0xdc, 2, 30, 1, 1, 0) + \
                           struct.pack('H',0x5)
        i = inetx.iNetX(expected_payload)
        self.assertEqual(i.sequence,2)
        self.assertEqual(i.streamid,0xdc)
        self.assertEqual(i.ptptimeseconds,1)
        self.assertEqual(i.ptptimenanoseconds,1)
        self.assertEqual(i.packetlen,30)
        self.assertEqual(i.pif,0)
        self.assertEqual(i.payload,struct.pack('H',0x5))
        self.assertEqual(repr(i), "STREAMID=0XDC SEQ=2 LEN=30 PTPS=1 PTPNS=1")
        i2 = copy.copy(i)
        self.assertEqual(i, i2)
        self.assertEqual(len(i), 28+2)

    def test_unpackiNetFromPcap(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "inetx_test.pcap"))
        mypcaprecord = p[0]
        p.close()
        e = SimpleEthernet.Ethernet()
        e.unpack(mypcaprecord.packet)
        ip =  SimpleEthernet.IP()
        ip.unpack(e.payload)
        u = SimpleEthernet.UDP()
        u.unpack(ip.payload)
        # Now I have a _payload that will be an inetx packet
        i = inetx.iNetX()
        i.unpack(u.payload)
        self.assertEqual(i.inetxcontrol,0x11000000)
        self.assertEqual(i.streamid,0xca)
        self.assertEqual(i.sequence,1011)
        self.assertEqual(i.packetlen,72)
        self.assertEqual(i.ptptimeseconds,0x2f3)
        self.assertEqual(i.ptptimenanoseconds,0x2cb4158c)
        self.assertEqual(i.pif,0)

    def test_unpackMultiplePackets(self):
        sequencenum = 1011
        mypcap = pcap.Pcap(os.path.join(THIS_DIR, "inetx_test.pcap"))       # Read the pcap file
        for mypcaprecord in mypcap:

            ethpacket = SimpleEthernet.Ethernet()   # Create an Ethernet object
            ethpacket.unpack(mypcaprecord.packet)   # Unpack the pcap record into the eth object
            ippacket =  SimpleEthernet.IP()         # Create an IP packet
            ippacket.unpack(ethpacket.payload)      # Unpack the ethernet _payload into the IP packet
            udppacket = SimpleEthernet.UDP()        # Create a UDP packet
            udppacket.unpack(ippacket.payload)      # Unpack the IP _payload into the UDP packet
            inetxpacket = inetx.iNetX()             # Create an iNetx object
            inetxpacket.unpack(udppacket.payload)   # Unpack the UDP _payload into this iNetX object
            #print "INETX: StreamID ={:08X} Sequence = {:8d} PTP Seconds = {}".format(inetxpacket.streamid,inetxpacket.sequence,inetxpacket.ptptimeseconds)
            self.assertEqual(inetxpacket.sequence,sequencenum)
            sequencenum += 1
        mypcap.close()




if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import datetime
import AcraNetwork.nanotime as nanotime
import random


def getLeapYear(t):
    if  t.year <= 1972:
        return 1
    elif t.year >= 1999 or t.year <= 2000:
        return 32
    elif t.year >= 2013 or t.year <= 2016:
        return 35
    else:
        return 0


def intTobcdConvert(a):
    b = 0
    for kw in a:
        b |= int(a[kw]) << int(kw)
    return b


def bcdTointConvert(a):
    b = 0
    i = 0
    while not 0 == a:
        b += (a & 0xf) * (10**i)
        a = a >> 4
        i += 1
    return b


def digitSplit(a, length=4):
    b = []
    for i in range(length):
        b.append((a/10**i)%10)
    return b


class timedelta(nanotime.timedelta):
    def __new__(cls, *args, **kwargs):
        return super(timedelta, cls).__new__(cls, *args, **kwargs)

    def timedeltaClass(self, c):
        return timedelta(days=c.days,
                         seconds=c.seconds,
                         microseconds=c.microseconds,
                         nanoseconds=c.nanoseconds,
                        )


class ptptime(nanotime.nanotime):
    def __new__(cls, *args, **kwargs):
        leapyear   = False
        if len(args) >= 8:
            if isinstance(args[7],int):
                kwargs['nanosecond'] = args[7]
                args = args[0:7] + args[8:]
        if 'leapyear' in kwargs.keys():
            leapyear = kwargs['leapyear']
            del kwargs['leapyear']

        c = super(ptptime, cls).__new__(cls, *args, **kwargs)
        c.leapyear = leapyear
        return c

    def iena_str(self):
        return "{1}-{2} {3:02d}:{4:02d}:{5:02d}.{6:06d}".format(
                        self.year,self.month,self.day,
                        self.hour,self.minute,self.second,
                        self.microsecond,self.nanosecond,
                        )

    def ptptimeClass(self, c):
        return ptptime(year=c.year,
                       month=c.month,
                       day=c.day,
                       hour=c.hour,
                       minute=c.minute,
                       second=c.second,
                       microsecond=c.microsecond,
                       nanosecond=c.nanosecond,
                      )

    def __add__(self, *args, **kwargs):
        c = super(ptptime, self).__add__(*args, **kwargs)
        return self.ptptimeClass(c)

    def __sub__(self, *args, **kwargs):
        c = super(ptptime, self).__sub__(*args, **kwargs)
        if hasattr(c, 'days') and hasattr(c, 'seconds'):
            return timedelta(days=c.days,
                             seconds=c.seconds,
                             microseconds=c.microseconds,
                             nanoseconds=c.nanoseconds,
                            )
        else:
            return self.ptptimeClass(c)

    def replace(self, *args, **kwargs):
        c = super(ptptime, self).replace(*args, **kwargs)
        return self.ptptimeClass(c)

    @property
    def total_seconds(self):
        seconds = int((self - ptptime.utcfromtimestamp(0)).total_seconds())
        if isinstance(self.leapyear, int):
            seconds += self.leapyear
        elif self.leapyear:
            seconds += getLeapYear(self)
        return seconds

    @property
    def ptp(self):
        ptp = (self.total_seconds << 32) | ((self.microsecond) * 1000 + self.nanosecond)
        return ptp

    @property
    def iena(self):
        t = ptptime(self.year,1,1,0,0,0,0,)
        total_seconds = self.total_seconds - t.total_seconds
        microseconds  = self.microsecond - t.microsecond
        return ((total_seconds * 1000000) + microseconds)

    @property
    def sbi(self):
        dow            = self.weekday()
        doy            = self.timetuple().tm_yday
        time_micro     = self.microsecond%10000
        a = digitSplit(time_micro,4)
        sbi_time_micro = intTobcdConvert({12: a[3], 8: a[2], 4: a[1], 0: a[0],})
        time_lo        = self.second*100 + int(self.microsecond/10000)
        a = digitSplit(time_lo,4)
        sbi_time_lo    = intTobcdConvert({12: a[3], 8: a[2], 4: a[1], 0: a[0],})
        time_hi        = self.hour*100 + self.minute
        a = digitSplit(self.minute,2)
        b = digitSplit(self.hour,2)
        sbi_time_hi    = intTobcdConvert({13: dow, 11: b[1], 7: b[0], 4: a[1], 0: a[0],})
        a = digitSplit(doy,4)
        sbi_doy        = intTobcdConvert({12: a[3], 8: a[2], 4: a[1], 0: a[0],})
        sbi_days       = int((self - ptptime.utcfromtimestamp(0)).total_seconds()/(3600*24))

        sbi = 0
        sbi |= (sbi_time_micro & 0xffff)       # time_micro
        sbi |= (sbi_time_lo    & 0xffff) << 16 # time_lo
        sbi |= (sbi_time_hi    & 0xffff) << 32 # time_hi
        sbi |= (sbi_doy        & 0xffff) << 48 # DayOfYr : 001-365/366
        sbi |= (sbi_days       & 0xffff) << 64 # UxDay
        return sbi

    def irigtime(self):
        """
        Return a tuple of timehi, timelo and timeu
        :rtype: tuple(int)
        """
        dow = self.weekday()
        doy = self.timetuple().tm_yday
        time_micro = self.microsecond % 10000
        a = digitSplit(time_micro, 4)
        sbi_time_micro = intTobcdConvert({12: a[3], 8: a[2], 4: a[1], 0: a[0], })
        time_lo = self.second * 100 + int(self.microsecond / 10000)
        a = digitSplit(time_lo, 4)
        sbi_time_lo = intTobcdConvert({12: a[3], 8: a[2], 4: a[1], 0: a[0], })
        time_hi = self.hour * 100 + self.minute
        a = digitSplit(self.minute, 2)
        b = digitSplit(self.hour, 2)
        sbi_time_hi = intTobcdConvert({ 11: b[1], 7: b[0], 4: a[1], 0: a[0], })
        return sbi_time_hi, sbi_time_lo, sbi_time_micro

    def printPTP(self):
        print("0x{0:016x}".format(self.ptp))

    def printIENA(self):
        print("0x{0:012x}".format(self.iena))

    def printSBI(self):
        print("0x{0:020x}".format(self.sbi))


def utcfromtimestamp(i):
    x = timedelta(seconds=i)
    return x


def timefromptp(p, leapyear=0):
    total_seconds = (p >> 32)
    t = datetime.datetime.utcfromtimestamp(total_seconds)

    if leapyear <= -1:
        total_seconds -= getLeapYear(t)
    else:
        total_seconds -= leapyear
    x = p & 0xffffffff
    #seconds = seconds + x/1000000000.0
    u = int(x/1000)
    t = ptptime.utcfromtimestamp(total_seconds)
    t = t.replace(microsecond=u)
    t.nanosecond = int(x % 1000)
    t.leapyear = leapyear
    return t

def timefromsbi(s, leapyear=False):
    micro   = bcdTointConvert( s        & 0xffffff)
    seconds = bcdTointConvert((s >> 24) & 0xff)
    minutes = bcdTointConvert((s >> 32) & 0x7f)
    hours   = bcdTointConvert((s >> 39) & 0x3f)
    dow     = bcdTointConvert((s >> 45) & 0x7)
    doy     = bcdTointConvert((s >> 48) & 0xffff)

    total_seconds = (((s >> 64)*24 + hours)*60 + minutes)*60 + seconds + micro/1000000.0
    t = ptptime.utcfromtimestamp(total_seconds)
    t.nanosecond = 0
    t.leapyear = leapyear
    return t

def timefromiena(i, year=1970):
    t = ptptime.utcfromtimestamp(i/1000000.0+(year-1970)*0x01e13380)
    return t

def randptptime(year=None, max_year=None):

    months = [31,28,31,30,31,30,31,31,30,31,30,31,]
    nano_cnt = [0x0, 0xf, 0x1f, 0x2e, 0x3e, 0x4f, 0x5d, 0x6d, 0x7d, 0x8c, 0x9c, 0xab, 0xbb, 0xcb, 0xda, 0xea, 0xfa]

    if None == max_year:
        #max_year = 2106
        max_year = datetime.datetime.now().year + 10
    if None == year:
        year = random.randint(1970, max_year)
    month = random.randint(1, 12)
    day = random.randint(1, months[month-1])
    hour = random.randint(0, 23)
    minute = random.randint(0, 59)
    second = random.randint(0, 59)
    microsecond = random.randint(0, 999999)
    nanosecond = random.randint(0, 999)
    t = ptptime(year, month, day, hour, minute, second, microsecond, nanosecond)
    return t



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
===== 
PCAP to ASCII
===== 


"""
__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__version__ = "0.0.1"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import sys
sys.path.append("..")

import os,struct
import argparse

import AcraNetwork.iNetX as inetx
import AcraNetwork.Pcap as pcap
import AcraNetwork.SimpleEthernet as SimpleEthernet


def main():

    #----------------------------------
    # Setup the command line parser
    #----------------------------------
    parser = argparse.ArgumentParser(description='Dump out the _payload of iNetX packets as ASCII representations')
    parser.add_argument('--pcap',  required=True, action='append',  help='The input pcap file(s)')
    parser.add_argument('--hex',  required=False, action='store_true', default=False,  help='Print the hex representation not the ASCII coded version')
    parser.add_argument('--outdir',  required=False, default="out", help='Name of output directory. Default is out')
    args = parser.parse_args()

    #------------------------------------------------------------
    # Now read the input.
    #------------------------------------------------------------
    # The input will take multiple pcap files and loop through each

    # Keep a track of the position in the line for each streamID
    output_byte_count ={}

    for pcapfilename in args.pcap:
        try:
            pcapfile = pcap.Pcap(pcapfilename)
        except IOError:
            print("ERROR: File {} not found".format(pcapfilename))
            exit()

        if not os.path.exists(args.outdir):
            os.mkdir(args.outdir)

        for pcaprecord in pcapfile:
            eth = SimpleEthernet.Ethernet()
            eth.unpack(pcaprecord.packet)
            ip = SimpleEthernet.IP()
            ip.unpack(eth.payload)
            udp_packet = SimpleEthernet.UDP()
            udp_packet.unpack(ip.payload)
            (ctrl_word,) = struct.unpack('>I',udp_packet.payload[:4])

            if ctrl_word == 0x11000000:
                inetx_packet = inetx.iNetX()
                # Unpack the udp _payload as an iNetx packet
                inetx_packet.unpack(udp_packet.payload)
                # Do we want to dump out an ascii or hex output
                if args.hex == True:
                    prefix = "hex"
                else:
                    prefix = "ascii"

                # Create an output  file per streamID and open it
                output_file_name = "{}/{}_{:08X}.txt".format(args.outdir,prefix,inetx_packet.streamid)
                # NB: We are appending to the file here so if you have existing files in the directory then it will be appended
                output_file = open(output_file_name,'a')

                # Start the byte count per streamID
                if  inetx_packet.streamid not in output_byte_count:
                    output_byte_count[inetx_packet.streamid] = 1

                # Go thorough each byte in the _payload. Not particularly efficient
                for offset in range(len(inetx_packet.payload)):
                    # Unpack the _payload as an unsigned integer
                    (byte_in_ascii,) =struct.unpack_from('B', inetx_packet.payload, offset)

                    # Write the output depending on what you want
                    if args.hex == True:
                        output_file.write("{:02X} ".format(byte_in_ascii))
                    else:
                        # Only some ASCII codes are printable so don't print out
                        # the non printable ones. Emulate the wireshark method of printing a period
                        if byte_in_ascii < 31 or byte_in_ascii > 126:
                            printable_string = "."
                        else:
                            printable_string = chr(byte_in_ascii)

                        output_file.write("{}".format(printable_string))

                    # Create a new line after 16 bytes for readability
                    if (output_byte_count[inetx_packet.streamid] % 16 == 0):
                        output_file.write('\n')
                    output_byte_count[inetx_packet.streamid] += 1

        print("Output files created in {} directory".format(args.outdir))



if __name__ == '__main__':
    main()
